from collections import defaultdict
from copy import deepcopy
from datetime import datetime
from functools import cached_property

from tabulate import tabulate

import vigorish.database as db
from vigorish.cli.components.viewers import create_display_table, create_table_viewer, DisplayTable
from vigorish.data.player_data import PlayerData
from vigorish.enums import DataSet, DefensePosition, VigFile
from vigorish.util.dt_format_strings import DT_AWARE, DT_AWARE_VERBOSE
from vigorish.util.exceptions import ScrapedDataException
from vigorish.util.list_helpers import flatten_list2d, group_and_sort_dict_list, make_chunked_list
from vigorish.util.regex import INNING_LABEL_REGEX
from vigorish.util.result import Result
from vigorish.util.string_helpers import (
    get_bbref_team_id,
    inning_number_to_string,
    validate_bbref_game_id,
    wrap_text,
)

AT_BAT_TABLE_MAX_ROWS = 8
AT_BAT_DESC_MAX_WIDTH = 35


class GameData:
    def __init__(self, app, bbref_game_id):

        # TODO: Create views for pitch_type by handedness by year
        # TODO: Create views for batter/pitcher stats to date, career, etc
        # TODO: Create view for invalid pitchfx
        # TODO: Create view to demonstrate how removed pfx is duplicative

        self.app = app
        self.db_engine = app.db_engine
        self.db_session = app.db_session
        self.scraped_data = app.scraped_data
        combined_data = self.scraped_data.get_combined_game_data(bbref_game_id)
        if not combined_data:
            raise ScrapedDataException(file_type=VigFile.COMBINED_GAME_DATA, data_set=DataSet.ALL, url_id=bbref_game_id)
        self.last_modified = combined_data["last_modified"]
        self.bbref_game_id = combined_data["bbref_game_id"]
        self.pitchfx_vs_bbref_audit = combined_data["pitchfx_vs_bbref_audit"]
        self.game_meta_info = combined_data["game_meta_info"]
        self.away_team_data = combined_data["away_team_data"]
        self.home_team_data = combined_data["home_team_data"]
        self.innings_list = combined_data["play_by_play_data"]
        self.removed_pitchfx = combined_data["removed_pitchfx"]
        self.invalid_pitchfx = combined_data["invalid_pitchfx"]
        self.player_id_dict = combined_data["player_id_dict"]
        self.all_player_mlb_ids = [id_map["mlb_id"] for id_map in self.player_id_dict.values()]
        self.game_date = validate_bbref_game_id(self.bbref_game_id).value["game_date"]
        self.player_data_cache = {}

    @property
    def away_team_id(self):
        return self.away_team_data["team_id_br"]

    @property
    def home_team_id(self):
        return self.home_team_data["team_id_br"]

    @property
    def away_team(self):
        return db.Team.find_by_team_id_and_year(self.db_session, self.away_team_id, self.game_date.year)

    @property
    def home_team(self):
        return db.Team.find_by_team_id_and_year(self.db_session, self.home_team_id, self.game_date.year)

    @property
    def game_datetime(self):
        return datetime.strptime(self.game_meta_info["game_date_time_str"], DT_AWARE)

    @property
    def team_name(self):
        return {
            self.away_team_id: self.away_team.name,
            self.home_team_id: self.home_team.name,
        }

    @property
    def team_data(self):
        return {
            self.away_team_id: self.away_team_data,
            self.home_team_id: self.home_team_data,
        }

    @property
    def all_at_bats(self):
        return self.valid_at_bats + self.invalid_at_bats

    @property
    def at_bat_map(self):
        return {at_bat["at_bat_id"]: at_bat for at_bat in self.all_at_bats}

    @property
    def pitch_stats_map(self):
        return {stats["pitcher_id_mlb"]: stats for stats in self.all_pitch_stats}

    @property
    def bat_stats_map(self):
        return {stats["batter_id_mlb"]: stats for stats in self.all_bat_stats}

    @property
    def bat_stats_player_ids(self):
        return list(self.bat_stats_map.keys())

    @property
    def pitch_stats_player_ids(self):
        return list(self.pitch_stats_map.keys())

    @property
    def winning_pitcher(self):
        pitcher_id_bbref = (
            self.away_team_data["pitcher_of_record"]
            if self.away_team_data["team_won"]
            else self.home_team_data["pitcher_of_record"]
        )
        return self.get_player_id_map(bbref_id=pitcher_id_bbref).mlb_id

    @property
    def losing_pitcher(self):
        pitcher_id_bbref = (
            self.away_team_data["pitcher_of_record"]
            if self.home_team_data["team_won"]
            else self.home_team_data["pitcher_of_record"]
        )
        return self.get_player_id_map(bbref_id=pitcher_id_bbref).mlb_id

    @property
    def pitcher_earned_save(self):
        pitcher_id_bbref = (
            self.away_team_data["pitcher_earned_save"]
            if self.away_team_data["team_won"] and self.away_team_data["pitcher_earned_save"]
            else self.home_team_data["pitcher_earned_save"]
            if self.home_team_data["team_won"] and self.home_team_data["pitcher_earned_save"]
            else None
        )
        return self.get_player_id_map(bbref_id=pitcher_id_bbref).mlb_id if pitcher_id_bbref else None

    @cached_property
    def valid_at_bats(self):
        return [at_bat_data for inning_data in self.innings_list for at_bat_data in inning_data["inning_events"]]

    @cached_property
    def invalid_at_bats(self):
        return [at_bat_data for inning_dict in self.invalid_pitchfx.values() for at_bat_data in inning_dict.values()]

    @cached_property
    def player_substitutions(self):
        return [
            event
            for at_bat in self.valid_at_bats
            for event in at_bat["pbp_events"]
            if event["event_type"] == "SUBSTITUTION"
        ]

    @cached_property
    def all_pitch_stats(self):
        return self.away_team_data["pitching_stats"] + self.home_team_data["pitching_stats"]

    @cached_property
    def all_bat_stats(self):
        bat_stats = self.away_team_data["batting_stats"] + self.home_team_data["batting_stats"]
        return [bat_stat for bat_stat in bat_stats if bat_stat["total_plate_appearances"]]

    @cached_property
    def bat_boxscore(self):
        return {
            self.away_team_id: self.create_bat_boxscore(self.away_team_id),
            self.home_team_id: self.create_bat_boxscore(self.home_team_id),
        }

    @cached_property
    def pitch_boxscore(self):
        return {
            self.away_team_id: self.create_pitch_boxscore(self.away_team_id),
            self.home_team_id: self.create_pitch_boxscore(self.home_team_id),
        }

    @cached_property
    def all_pitchfx(self):
        pfx_dict = defaultdict(list)
        all_pitchfx = [at_bat["pitchfx"] for at_bat in self.all_at_bats]
        all_pitchfx.extend(self.get_removed_pfx())
        for pfx in flatten_list2d(all_pitchfx):
            pfx_dict[pfx["pitch_app_id"]].append(self._update_pfx_attributes(pfx))
        return pfx_dict

    @cached_property
    def starting_pitcher_mlb_ids(self):
        return [
            mlb_id
            for mlb_id, pitch_stats in self.pitch_stats_map.items()
            if check_pitch_stats_for_game_score(pitch_stats)
        ]

    @cached_property
    def extra_innings(self):
        return len(self.get_inning_numbers_sorted()) > 9

    def get_game_data(self):
        return {
            "game_id": self.bbref_game_id,
            "linescore": self.get_html_linescore(),
            "pitcher_results": self.get_pitcher_results(),
            "extra_innings": self.extra_innings,
        }

    def get_boxscore_data(self):
        boxscore_data = {
            "last_modified": self.last_modified,
            "game_id": self.bbref_game_id,
            "away_team": self.get_team_data(self.away_team_id),
            "home_team": self.get_team_data(self.home_team_id),
            "extra_innings": self.extra_innings,
            "game_meta": self.get_game_meta_info(),
            "linescore": self.get_html_linescore(),
        }
        if self.extra_innings:
            boxscore_data["linescore_complete"] = self.get_html_linescore(condensed=False)
        return boxscore_data

    def get_team_data(self, team_id):
        (team_bat_boxscore, team_pitch_boxscore) = self._prepare_team_data(team_id)
        return {
            "team_id": team_id,
            "team_name": self.team_name[team_id],
            "total_wins": self.team_data[team_id]["total_wins_before_game"],
            "total_losses": self.team_data[team_id]["total_losses_before_game"],
            "runs_scored_by_team": self.team_data[team_id]["total_runs_scored_by_team"],
            "runs_scored_by_opponent": self.team_data[team_id]["total_runs_scored_by_opponent"],
            "team_won": self.team_data[team_id]["team_won"],
            "pitcher_of_record": self.team_data[team_id]["pitcher_of_record"],
            "pitcher_earned_save": self.team_data[team_id]["pitcher_earned_save"],
            "batting": team_bat_boxscore,
            "pitching": team_pitch_boxscore,
        }

    def _prepare_team_data(self, team_id):
        team_data = deepcopy(self.team_data[team_id])
        team_data.pop("starting_lineup")
        bat_map = {b["batter_id_mlb"]: b for b in team_data.pop("batting_stats")}
        bat_boxscore = self.bat_boxscore[team_id]
        for box in bat_boxscore.values():
            box["def_position"] = int(box["def_position"])
            box["total_pbp_events"] = bat_map[box["mlb_id"]]["total_pbp_events"]
            box["total_incomplete_at_bats"] = bat_map[box["mlb_id"]]["total_incomplete_at_bats"]
            box["total_plate_appearances"] = bat_map[box["mlb_id"]]["total_plate_appearances"]
            box["at_bat_results"] = self._summarize_all_at_bats_for_player(box["mlb_id"])
            box["incomplete_at_bat_ids"] = bat_map[box["mlb_id"]]["incomplete_at_bat_ids"]
            box["substitutions"] = self._find_player_sub_events(box["mlb_id"])
            box["bbref_data"] = bat_map[box["mlb_id"]]["bbref_data"]
        pitch_map = {p["pitcher_id_mlb"]: p for p in team_data.pop("pitching_stats")}
        pitch_boxscore = self.pitch_boxscore[team_id]
        for box in pitch_boxscore.values():
            box["pitch_app_id"] = pitch_map[box["mlb_id"]]["pitch_app_id"]
            box["pitch_count_by_inning"] = pitch_map[box["mlb_id"]]["pitch_count_by_inning"]
            box["pitch_count_by_pitch_type"] = pitch_map[box["mlb_id"]]["pitch_count_by_pitch_type"]
            box["at_bat_ids"] = pitch_map[box["mlb_id"]]["pitch_app_pitchfx_audit"]["at_bat_ids_pitchfx_complete"]
            box["substitutions"] = self._find_player_sub_events(box["mlb_id"])
            box["bbref_data"] = pitch_map[box["mlb_id"]]["bbref_data"]
            box["inning_totals"] = self._summarize_pitch_app(box["mlb_id"])
        return (bat_boxscore, pitch_boxscore)

    def _summarize_all_at_bats_for_player(self, mlb_id):
        return [self._summarize_at_bat(ab) for ab in self.valid_at_bats if ab["batter_id_mlb"] == mlb_id]

    def _summarize_at_bat(self, at_bat):
        pbp_events = [event for event in at_bat["pbp_events"] if event["event_type"] == "AT_BAT"]
        pbp_events.sort(key=lambda x: x["pbp_table_row_number"])
        return {
            "at_bat_id": at_bat["at_bat_id"],
            "pbp_table_row_number": at_bat["pbp_table_row_number"],
            "batter_name": at_bat["batter_name"],
            "pitcher_name": at_bat["pitcher_name"],
            "inning": pbp_events[-1]["inning_label"],
            "runners_on_base": pbp_events[-1]["runners_on_base"],
            "outs": pbp_events[-1]["outs_before_play"],
            "play_description": pbp_events[-1]["play_description"],
            "pitch_sequence": pbp_events[-1]["pitch_sequence"],
        }

    def _find_player_sub_events(self, mlb_id):
        player_id = self.get_player_id_map(mlb_id=mlb_id)
        bbref_id = player_id.bbref_id
        return [
            sub_event
            for sub_event in self.player_substitutions
            if bbref_id in [sub_event["incoming_player_id_br"], sub_event["outgoing_player_id_br"]]
        ]

    def _summarize_pitch_app(self, mlb_id):
        pfx = self.get_pfx_for_pitcher(mlb_id).value
        pfx_by_inning = group_and_sort_dict_list(pfx, "inning", "pitch_id")
        at_bats = [ab for ab in self.valid_at_bats if ab["pitcher_id_mlb"] == mlb_id]
        at_bats_by_inning = group_and_sort_dict_list(at_bats, "inning_id", "pbp_table_row_number")
        inning_totals = []
        for (inning_pfx, inning_at_bats) in zip(pfx_by_inning.values(), at_bats_by_inning.values()):
            totals = {
                "outs": sum(ab["runs_outs_result"].count("O") for ab in inning_at_bats),
                "hits": sum(pfx["ab_result_hit"] for pfx in inning_pfx),
                "runs": sum(ab["runs_outs_result"].count("R") for ab in inning_at_bats),
                "bb": sum(pfx["ab_result_bb"] for pfx in inning_pfx),
                "so": sum(pfx["ab_result_k"] for pfx in inning_pfx),
                "bf": len({pfx["batter_id_mlb"] for pfx in inning_pfx}),
                "pitch_count": len(inning_pfx),
                "strikes": len([pfx for pfx in inning_pfx if pfx["basic_type"] in ["S", "X"]]),
            }
            inning_totals.append(totals)
        return dict(zip(list(pfx_by_inning.keys()), inning_totals))

    def get_pfx_for_pitcher(self, mlb_id):
        result = self.validate_mlb_id(mlb_id)
        if result.failure:
            return result
        mlb_id = result.value
        pfx = [at_bat["pitchfx"] for at_bat in self.valid_at_bats if at_bat["pitcher_id_mlb"] == mlb_id]
        return Result.Ok(self._prepare_pfx_for_api_response(flatten_list2d(pfx)))

    def get_pfx_for_at_bat(self, at_bat_id):
        if at_bat_id not in self.at_bat_map:
            return None
        pfx = self.at_bat_map[at_bat_id]["pitchfx"]
        return self._prepare_pfx_for_api_response(pfx)

    def _prepare_pfx_for_api_response(self, pfx):
        pfx = list(map(db.PitchFx.update_pfx_dict, deepcopy(pfx)))
        pfx = list(map(self._update_pfx_attributes, pfx))
        return sorted(pfx, key=lambda x: x["time_pitch_thrown_utc"])

    def _update_pfx_attributes(self, pfx):
        pfx["is_sp"] = pfx["pitcher_id"] in self.starting_pitcher_mlb_ids
        pfx["is_rp"] = pfx["pitcher_id"] not in self.starting_pitcher_mlb_ids
        return pfx

    def get_player_data(self, mlb_id):
        player_data = self.player_data_cache.get(mlb_id)
        if not player_data:
            player_data = PlayerData(self.app, mlb_id)
            self.player_data_cache[mlb_id] = player_data
        return player_data

    def create_bat_boxscore(self, team_id):
        team_data = self.team_data[team_id]
        batter_box = {}
        for slot in team_data["starting_lineup"]:
            mlb_id = self.get_player_id_map(bbref_id=slot["player_id_br"]).mlb_id
            bat_order = slot["bat_order"]
            def_pos = slot["def_position"]
            batter_box[slot["bat_order"]] = self.get_bat_boxscore_for_player(mlb_id, bat_order, def_pos, team_data)
        lineup_player_ids = [bat_boxscore["mlb_id"] for bat_boxscore in batter_box.values()]
        sub_player_ids = [
            mlb_id
            for mlb_id in self.get_all_player_ids_by_team(team_data["team_id_br"])
            if mlb_id in self.bat_stats_player_ids and mlb_id not in lineup_player_ids
        ]
        for num, sub_id in enumerate(sub_player_ids, start=1):
            batter_box[f"BN{num}"] = self.get_bat_boxscore_for_player(sub_id, 0, "BN", team_data)
        return batter_box

    def get_player_id_map(self, mlb_id=None, bbref_id=None):
        if not mlb_id and not bbref_id:
            return None
        return (
            db.PlayerId.find_by_mlb_id(self.db_session, mlb_id)
            if mlb_id
            else db.PlayerId.find_by_bbref_id(self.db_session, bbref_id)
        )

    def get_all_player_ids_by_team(self, team_id):
        return [
            player_dict["mlb_id"]
            for player_dict in self.player_id_dict.values()
            if player_dict["team_id_bbref"] == team_id
        ]

    def get_bat_boxscore_for_player(self, mlb_id, bat_order, def_pos, team_data):
        player_id = self.get_player_id_map(mlb_id=mlb_id)
        bat_stats = self.get_bat_stats(mlb_id).value
        (at_bats, details) = parse_bat_stats_for_game(bat_stats)
        def_position = DefensePosition.from_abbrev(def_pos)
        return {
            "team_id": team_data["team_id_br"],
            "name": player_id.mlb_name,
            "mlb_id": player_id.mlb_id,
            "bbref_id": player_id.bbref_id,
            "is_starter": def_position.is_starter,
            "bat_order": bat_order if def_position.is_starter else 0,
            "def_position": def_position,
            "at_bats": at_bats,
            "bat_stats": details,
            "stats_to_date": parse_bat_stats_to_date(bat_stats),
        }

    def create_pitch_boxscore(self, team_id):
        team_data = self.team_data[team_id]
        pitcher_box = {}
        pitcher_ids = [
            mlb_id
            for mlb_id in self.get_all_player_ids_by_team(team_data["team_id_br"])
            if mlb_id in self.pitch_stats_player_ids
        ]
        rp_count = 0
        for mlb_id in self.get_pitcher_app_order(pitcher_ids):
            player_box = self.get_pitch_boxscore_for_player(mlb_id, team_data)
            if player_box["pitch_app_type"] == "SP":
                pitcher_box["SP"] = player_box
            else:
                rp_count += 1
                pitcher_box[f"RP{rp_count}"] = player_box
        return pitcher_box

    def get_pitcher_app_order(self, pitcher_ids):
        pitcher_app_dict = {
            mlb_id: min(at_bat["pbp_table_row_number"] for at_bat in self.get_valid_at_bats_for_pitcher(mlb_id).value)
            for mlb_id in pitcher_ids
        }
        return sorted(pitcher_ids, key=lambda x: pitcher_app_dict[x])

    def get_pitch_boxscore_for_player(self, mlb_id, team_data):
        player_id = self.get_player_id_map(mlb_id=mlb_id)
        pitch_stats = self.get_pitch_app_stats(mlb_id).value
        game_stats = parse_pitch_app_stats(pitch_stats)
        is_starter = mlb_id in self.starting_pitcher_mlb_ids
        return {
            "team_id": team_data["team_id_br"],
            "name": player_id.mlb_name,
            "mlb_id": mlb_id,
            "bbref_id": player_id.bbref_id,
            "pitch_app_type": "SP" if is_starter else "RP",
            "game_results": game_stats,
        }

    def view_valid_at_bats_for_batter(self, mlb_id):
        result = self.validate_mlb_id(mlb_id)
        if result.failure:
            return result
        mlb_id = result.value
        at_bats = self.get_valid_at_bats_for_batter(mlb_id).value
        batter_tables = []
        for num, at_bat in enumerate(at_bats, start=1):
            player_name = self.get_player_id_map(mlb_id=mlb_id).mlb_name
            heading = f"At Bat #{num}/{len(at_bats)} for {player_name} in Game {self.bbref_game_id}"
            table_list = self.create_at_bat_table_list(at_bat["at_bat_id"], heading)
            batter_tables.extend(table_list)
        table_viewer = create_table_viewer(batter_tables)
        return Result.Ok(table_viewer)

    def validate_mlb_id(self, mlb_id):
        if not isinstance(mlb_id, int):
            try:
                mlb_id = int(mlb_id)
            except ValueError:
                return Result.Fail(f'"{mlb_id}" could not be parsed as a valid int')
        if mlb_id not in self.all_player_mlb_ids:
            error = f"Error: Player MLB ID: {mlb_id} did not appear in game {self.bbref_game_id}"
            return Result.Fail(error)
        return Result.Ok(mlb_id)

    def get_valid_at_bats_for_batter(self, mlb_id):
        result = self.validate_mlb_id(mlb_id)
        if result.failure:
            return result
        mlb_id = result.value
        at_bats = [ab for ab in self.valid_at_bats if ab["batter_id_mlb"] == mlb_id]
        return Result.Ok(at_bats)

    def create_at_bat_table_list(self, at_bat_id, heading=None):
        at_bat = self.at_bat_map.get(at_bat_id)
        if not at_bat:
            return []
        message = self.get_at_bat_details(at_bat)
        pitch_seq_desc = format_pitch_sequence_description(at_bat["pitch_sequence_description"])
        chunked_list = make_chunked_list(pitch_seq_desc, chunk_size=AT_BAT_TABLE_MAX_ROWS)
        return [create_display_table(chunk, heading, message) for chunk in chunked_list]

    def get_at_bat_details(self, at_bat):
        pitch_app_stats = self.get_pitch_app_stats(at_bat["pitcher_id_mlb"]).value
        bat_stats = self.get_bat_stats(at_bat["batter_id_mlb"]).value
        return get_at_bat_details(at_bat, pitch_app_stats, bat_stats["bbref_data"])

    def get_pitch_app_stats(self, mlb_id):
        result = self.validate_mlb_id(mlb_id)
        if result.failure:
            return result
        mlb_id = result.value
        pitch_stats = self.pitch_stats_map.get(mlb_id, {})
        if pitch_stats:
            sp_mlb_ids = self.starting_pitcher_mlb_ids
            pitch_stats["is_sp"] = pitch_stats["pitcher_id_mlb"] in sp_mlb_ids
            pitch_stats["is_rp"] = pitch_stats["pitcher_id_mlb"] not in sp_mlb_ids
            pitch_stats["is_wp"] = pitch_stats["pitcher_id_mlb"] == self.winning_pitcher
            pitch_stats["is_lp"] = pitch_stats["pitcher_id_mlb"] == self.losing_pitcher
            pitch_stats["is_sv"] = pitch_stats["pitcher_id_mlb"] == self.pitcher_earned_save
        return Result.Ok(deepcopy(pitch_stats))

    def get_bat_stats(self, mlb_id):
        result = self.validate_mlb_id(mlb_id)
        if result.failure:
            return result
        mlb_id = result.value
        bat_stats = self.bat_stats_map.get(mlb_id, {})
        return Result.Ok(deepcopy(bat_stats))

    def view_valid_at_bats_for_pitcher(self, mlb_id):
        result = self.validate_mlb_id(mlb_id)
        if result.failure:
            return result
        mlb_id = result.value
        at_bats = self.get_valid_at_bats_for_pitcher(mlb_id).value
        at_bats_by_inning = group_and_sort_dict_list(at_bats, "inning_id", "pbp_table_row_number")
        at_bat_tables_by_inning = {}
        all_at_bat_tables = []
        for inning_id, at_bats in at_bats_by_inning.items():
            inning = inning_id[-5:]
            at_bat_tables = []
            for num, at_bat in enumerate(at_bats, start=1):
                player_name = self.get_player_id_map(mlb_id=mlb_id).mlb_name
                heading = f"Inning: {inning}, At Bat #{num}/{len(at_bats)}, P: {player_name}"
                table_list = self.create_at_bat_table_list(at_bat["at_bat_id"], heading)
                at_bat_tables.extend(table_list)
            at_bat_tables_by_inning[inning] = at_bat_tables
            all_at_bat_tables.extend(at_bat_tables)

        innings_viewer = {"ALL": create_table_viewer(all_at_bat_tables)}
        for inning, at_bat_tables in at_bat_tables_by_inning.items():
            innings_viewer[inning] = create_table_viewer(at_bat_tables)
        return Result.Ok(innings_viewer)

    def get_valid_at_bats_for_pitcher(self, mlb_id):
        result = self.validate_mlb_id(mlb_id)
        if result.failure:
            return result
        mlb_id = result.value
        at_bats = [at_bat for at_bat in self.valid_at_bats if at_bat["pitcher_id_mlb"] == mlb_id]
        return Result.Ok(at_bats)

    def view_at_bats_by_inning(self):
        all_at_bat_tables = []
        at_bat_tables_by_inning = {}
        for inning_dict in self.get_innings_sorted():
            inning = inning_dict["inning_id"][-5:]
            inning_tables = []
            for num, at_bat in enumerate(inning_dict["at_bats"], start=1):
                heading = f"Inning: {inning}, At Bat #{num}/{len(inning_dict['at_bats'])}"
                table_list = self.create_at_bat_table_list(at_bat["at_bat_id"], heading)
                inning_tables.extend(table_list)
            at_bat_tables_by_inning[inning] = inning_tables
            all_at_bat_tables.extend(inning_tables)

        innings_viewer = {"ALL": create_table_viewer(all_at_bat_tables)}
        for inning, at_bat_tables in at_bat_tables_by_inning.items():
            innings_viewer[inning] = create_table_viewer(at_bat_tables)
        return Result.Ok(innings_viewer)

    def get_innings_sorted(self):
        at_bats_by_inning = group_and_sort_dict_list(self.valid_at_bats, "inning_id", "pbp_table_row_number")
        innings_unsorted = [
            {"inning_id": inning_id, "at_bats": at_bats} for inning_id, at_bats in at_bats_by_inning.items()
        ]
        return get_innings_sorted(innings_unsorted)

    def view_game_meta_info(self):
        meta_info = self.get_game_meta_info()
        meta_info.pop("umpires")
        game_start_time = meta_info.pop("game_start_time")
        table_rows = [[name, value] for name, value in meta_info.items()]
        table_rows.insert(0, ["game_start_time", game_start_time])
        table = tabulate(table_rows)
        heading = f"Meta Information for game {self.bbref_game_id}"
        return create_table_viewer([DisplayTable(table, heading)])

    def get_game_meta_info(self):
        meta_info = deepcopy(self.game_meta_info)
        meta_info.pop("game_time_hour")
        meta_info.pop("game_time_minute")
        meta_info["game_start_time"] = meta_info.pop("game_date_time_str")
        return meta_info

    def get_removed_pfx(self):
        return [at_bat["pitchfx"] for inning_dict in self.removed_pitchfx.values() for at_bat in inning_dict.values()]

    def view_pitch_mix_batter_stance_splits(self, mlb_id):
        result = self.validate_mlb_id(mlb_id)
        if result.failure:
            return result
        mlb_id = result.value
        pitcher_data = self.get_player_data(mlb_id)
        table_viewer = pitcher_data.view_pitch_mix_batter_stance_splits(self.bbref_game_id)
        return Result.Ok(table_viewer)

    def view_pitch_mix_season_splits(self, mlb_id):
        result = self.validate_mlb_id(mlb_id)
        if result.failure:
            return result
        mlb_id = result.value
        pitcher_data = self.get_player_data(mlb_id)
        table_viewer = pitcher_data.view_pitch_mix_season_splits(self.bbref_game_id)
        return Result.Ok(table_viewer)

    def view_pd_pitch_type_splits_for_pitcher(self, mlb_id):
        result = self.validate_mlb_id(mlb_id)
        if result.failure:
            return result
        mlb_id = result.value
        pitcher_data = self.get_player_data(mlb_id)
        table_viewer = pitcher_data.view_plate_discipline_pitch_type_splits(self.bbref_game_id)
        return Result.Ok(table_viewer)

    def view_bb_pitch_type_splits_for_pitcher(self, mlb_id):
        result = self.validate_mlb_id(mlb_id)
        if result.failure:
            return result
        mlb_id = result.value
        pitcher_data = self.get_player_data(mlb_id)
        table_viewer = pitcher_data.view_batted_ball_pitch_type_splits(self.bbref_game_id)
        return Result.Ok(table_viewer)

    def view_bat_stats_pitch_type_splits_for_pitcher(self, mlb_id):
        result = self.validate_mlb_id(mlb_id)
        if result.failure:
            return result
        mlb_id = result.value
        pitcher_data = self.get_player_data(mlb_id)
        table_viewer = pitcher_data.view_bat_stats_pitch_type_splits(self.bbref_game_id)
        return Result.Ok(table_viewer)

    def get_matchup_details(self):
        away_record = get_team_record_for_linescore(self.away_team_data)
        home_record = get_team_record_for_linescore(self.home_team_data)
        matchup = f"{self.away_team.name} ({away_record}) vs {self.home_team.name} ({home_record})"
        game_time = self.game_datetime.strftime(DT_AWARE_VERBOSE)
        w_l_sv = self.get_pitcher_results_str()
        return f"{game_time}\n{matchup}\n{w_l_sv}\n"

    def get_pitcher_results_str(self):
        pitchers = self.get_pitcher_results()
        w_pitcher = f'WP: {pitchers["wp"]["name"]} ({pitchers["wp"]["team_id"]})'
        l_pitcher = f'LP: {pitchers["lp"]["name"]} ({pitchers["lp"]["team_id"]})'
        if "sv" not in pitchers:
            return f"{w_pitcher} {l_pitcher}"
        sv_pitcher = f'SV: {pitchers["sv"]["name"]} ({pitchers["wp"]["team_id"]})'
        return f"{w_pitcher} {l_pitcher} {sv_pitcher}"

    def get_pitcher_results(self):
        pitcher_results = {}
        w_pitcher_bbref_id = (
            self.away_team_data["pitcher_of_record"]
            if self.away_team_data["team_won"]
            else self.home_team_data["pitcher_of_record"]
        )
        w_pitcher_id = self.get_player_id_map(bbref_id=w_pitcher_bbref_id).mlb_id
        w_pitcher_name = self.get_player_id_map(bbref_id=w_pitcher_bbref_id).mlb_name
        w_pitcher_team_id = self.away_team_id if self.away_team_data["team_won"] else self.home_team_id
        pitcher_results["wp"] = {"mlb_id": w_pitcher_id, "name": w_pitcher_name, "team_id": w_pitcher_team_id}
        l_pitcher_bbref_id = (
            self.home_team_data["pitcher_of_record"]
            if self.away_team_data["team_won"]
            else self.away_team_data["pitcher_of_record"]
        )
        l_pitcher_id = self.get_player_id_map(bbref_id=l_pitcher_bbref_id).mlb_id
        l_pitcher_name = self.get_player_id_map(bbref_id=l_pitcher_bbref_id).mlb_name
        l_pitcher_team_id = self.home_team_id if self.away_team_data["team_won"] else self.away_team_id
        pitcher_results["lp"] = {"mlb_id": l_pitcher_id, "name": l_pitcher_name, "team_id": l_pitcher_team_id}
        sv_pitcher_bbref_id = (
            self.away_team_data["pitcher_earned_save"]
            if self.away_team_data["team_won"] and self.away_team_data["pitcher_earned_save"]
            else self.home_team_data["pitcher_earned_save"]
            if self.home_team_data["team_won"] and self.home_team_data["pitcher_earned_save"]
            else 0
        )
        if sv_pitcher_bbref_id:
            sv_pitcher_id = self.get_player_id_map(bbref_id=sv_pitcher_bbref_id).mlb_id
            sv_pitcher_name = self.get_player_id_map(bbref_id=sv_pitcher_bbref_id).mlb_name
            pitcher_results["sv"] = {"mlb_id": sv_pitcher_id, "name": sv_pitcher_name, "team_id": w_pitcher_team_id}
        return pitcher_results

    def get_tui_linescore(self):
        table_list = [
            create_tui_linescore(**html_linescore)
            for html_linescore in get_html_linescore(
                self.away_team_data, self.home_team_data, self.innings_list, chunkify=True
            )
        ]
        return "\n\n".join(table_list)

    def get_html_linescore(self, condensed=True):
        linescore_components = get_html_linescore(
            self.away_team_data, self.home_team_data, self.innings_list, condensed
        )
        if condensed and self.extra_innings:
            linescore_components = condense_html_linescore(linescore_components)
        return linescore_components

    def get_inning_numbers_sorted(self):
        inning_labels = [inning["inning_id"] for inning in self.get_innings_sorted()]
        return sorted({parse_inning_number_from_label(inning) for inning in inning_labels})

    def get_inning_runs_scored_map(self):
        html_linescore = get_html_linescore(self.away_team_data, self.home_team_data, self.innings_list, False)
        return get_inning_runs_scored_map(html_linescore)


def get_html_linescore(away_team_data, home_team_data, innings_list, condensed=False, chunkify=False):
    linescore_tables = []
    (total_chunks, away_team_runs_chunked, home_team_runs_chunked) = get_runs_by_inning(innings_list, chunkify)
    for i in range(total_chunks):
        start_inning = i * 9 + 1
        linescore = create_html_linescore_components(
            away_team_data["team_id_br"],
            away_team_runs_chunked[i],
            get_team_totals_for_linescore(away_team_data),
            home_team_data["team_id_br"],
            home_team_runs_chunked[i],
            get_team_totals_for_linescore(home_team_data),
            start_inning,
            is_last_chunk(total_chunks, i + 1),
            is_extra_innings_game(total_chunks, away_team_runs_chunked[i]),
        )
        if condensed:
            linescore.pop("last_chunk")
        linescore_tables.append(linescore)
    return linescore_tables


def condense_html_linescore(linescore_tables):
    inning_runs_scored_map = get_inning_runs_scored_map(linescore_tables)
    (condensed_linescore_map, removed_inning_map, inning_runs_scored_map) = _condense_linescore(inning_runs_scored_map)
    return {
        "inning_numbers": [x["inning"] for x in condensed_linescore_map],
        "game_totals": ["R", "H", "E"],
        "away_team_id": linescore_tables[0]["away_team_id"],
        "away_team_runs_by_inning": [x["away_runs"] for x in condensed_linescore_map],
        "away_team_totals": linescore_tables[0]["away_team_totals"],
        "home_team_id": linescore_tables[0]["home_team_id"],
        "home_team_runs_by_inning": [x["home_runs"] for x in condensed_linescore_map],
        "home_team_totals": linescore_tables[0]["home_team_totals"],
        "extra_innings": linescore_tables[0]["extra_innings"],
        "removed_innings": removed_inning_map,
        "inning_runs_scored_map": inning_runs_scored_map,
    }


def get_inning_runs_scored_map(linescore_tables):
    inning_runs_scored_map = []
    if isinstance(linescore_tables, list):
        for linescore in linescore_tables:
            inning_map = _get_inning_runs_scored_map(linescore)
            inning_runs_scored_map.extend(inning_map)
    else:
        inning_runs_scored_map = _get_inning_runs_scored_map(linescore_tables)
    (_, _, inning_runs_scored_map) = _condense_linescore(inning_runs_scored_map)
    return inning_runs_scored_map


def _get_inning_runs_scored_map(linescore):
    inning_runs_scored_map = []
    inning_numbers_without_filler = [num for num in linescore["inning_numbers"] if num != ""]
    for i in range(len(inning_numbers_without_filler)):
        inning = linescore["inning_numbers"][i]
        away_team_runs_scored = linescore["away_team_runs_by_inning"][i]
        home_team_runs_scored = linescore["home_team_runs_by_inning"][i]
        inning_somebody_scored = away_team_runs_scored != 0 or home_team_runs_scored != 0
        home_team_runs_int = 0 if isinstance(home_team_runs_scored, str) else home_team_runs_scored
        inning_runs_scored_map.append(
            {
                "runs_scored": inning_somebody_scored,
                "inning": inning,
                "away_runs": away_team_runs_scored,
                "home_runs": home_team_runs_scored,
                "total_runs_scored": away_team_runs_scored + home_team_runs_int,
            }
        )
    return inning_runs_scored_map


def _condense_linescore(inning_runs_scored_map):
    innings_nobody_scored = list(filter(lambda x: not x["runs_scored"], inning_runs_scored_map))
    innings_somebody_scored = list(filter(lambda x: x["runs_scored"], inning_runs_scored_map))
    removed_innings = []
    if len(innings_somebody_scored) < 9:
        missing_column_count = 9 - len(innings_somebody_scored)
        innings_somebody_scored.extend(innings_nobody_scored[:missing_column_count])
        removed_innings = [inn["inning"] for inn in innings_nobody_scored[missing_column_count:]]
    elif len(innings_somebody_scored) > 9:
        remove_column_count = len(innings_somebody_scored) - 9
        innings_somebody_scored.sort(key=lambda x: x["total_runs_scored"])
        innings_somebody_scored = innings_somebody_scored[remove_column_count:]
        removed_innings = [inn["inning"] for inn in innings_nobody_scored[:remove_column_count]]
    removed_inning_map = []
    for inning_map in inning_runs_scored_map:
        removed_inning_map.append(inning_map["inning"] in removed_innings)
        inning_map["removed_inning"] = inning_map["inning"] in removed_innings
    return (sorted(innings_somebody_scored, key=lambda x: x["inning"]), removed_inning_map, inning_runs_scored_map)


def get_runs_by_inning(innings_list, chunkify):
    is_top_half = True
    away_team_runs = []
    home_team_runs = []
    for inning in innings_list:
        total_runs_this_inning = get_total_runs_in_inning(inning["inning_events"])
        if is_top_half:
            away_team_runs.append(total_runs_this_inning)
        else:
            home_team_runs.append(total_runs_this_inning)
        is_top_half = not is_top_half
    if len(away_team_runs) > len(home_team_runs):
        home_team_runs.append("X")
    if not chunkify:
        return (1, [away_team_runs], [home_team_runs])
    away_team_runs_chunked = make_chunked_list(away_team_runs, chunk_size=9)
    home_team_runs_chunked = make_chunked_list(home_team_runs, chunk_size=9)
    total_nine_inning_chunks = len(away_team_runs_chunked)
    return (total_nine_inning_chunks, away_team_runs_chunked, home_team_runs_chunked)


def get_total_runs_in_inning(inning_at_bats):
    return sum(ab["runs_outs_result"].count("R") for ab in inning_at_bats)


def is_last_chunk(total_linescore_tables, current_table_number):
    return current_table_number == total_linescore_tables


def is_extra_innings_game(total_linescore_tables, innings_list):
    return total_linescore_tables > 1 or len(innings_list) > 9


def create_html_linescore_components(
    away_team_id,
    away_team_runs_by_inning,
    away_team_totals,
    home_team_id,
    home_team_runs_by_inning,
    home_team_totals,
    start_inning,
    last_chunk,
    extra_innings,
):
    end_inning = start_inning + len(away_team_runs_by_inning)
    inning_numbers = list(range(start_inning, end_inning))
    if last_chunk and extra_innings:
        away_team_runs_by_inning = add_filler_columns(away_team_runs_by_inning)
        home_team_runs_by_inning = add_filler_columns(home_team_runs_by_inning)
        inning_numbers = add_filler_columns(inning_numbers)

    return {
        "inning_numbers": inning_numbers,
        "game_totals": ["R", "H", "E"],
        "away_team_id": away_team_id,
        "away_team_runs_by_inning": away_team_runs_by_inning,
        "away_team_totals": away_team_totals,
        "home_team_id": home_team_id,
        "home_team_runs_by_inning": home_team_runs_by_inning,
        "home_team_totals": home_team_totals,
        "last_chunk": last_chunk,
        "extra_innings": extra_innings,
    }


def create_tui_linescore(
    inning_numbers,
    game_totals,
    away_team_id,
    away_team_runs_by_inning,
    away_team_totals,
    home_team_id,
    home_team_runs_by_inning,
    home_team_totals,
    last_chunk,
    extra_innings,
):
    inning_numbers = [get_tui_inning_number(num) for num in inning_numbers]
    tui_inning_numbers = [f"{' '*(4-len(str(num)))}{num}" for num in inning_numbers]
    tui_game_totals = [f"{' '*(4-len(letter))}{letter}" for letter in game_totals]
    tui_away_inning_runs = [f"{' '*(4-len(str(num)))}{num}" for num in away_team_runs_by_inning]
    tui_away_totals = [f"{' '*(4-len(str(num)))}{num}" for num in away_team_totals]
    tui_home_inning_runs = [f"{' '*(4-len(str(num)))}{num}" for num in home_team_runs_by_inning]
    tui_home_totals = [f"{' '*(4-len(str(num)))}{num}" for num in home_team_totals]
    return format_tui_linescore_components(
        tui_inning_numbers,
        tui_game_totals,
        away_team_id,
        tui_away_inning_runs,
        tui_away_totals,
        home_team_id,
        tui_home_inning_runs,
        tui_home_totals,
        last_chunk,
    )


def format_tui_linescore_components(
    tui_inning_numbers,
    tui_game_totals,
    away_team_id,
    tui_away_inning_runs,
    tui_away_totals,
    home_team_id,
    tui_home_inning_runs,
    tui_home_totals,
    last_chunk,
):
    linescore = f'   {"".join(tui_inning_numbers)}'
    if last_chunk:
        linescore += "".join(tui_game_totals)
    linescore += f"\n----{'----'*9}"
    if last_chunk:
        linescore += f"{'----'*3}"
    linescore += f"\n{away_team_id}{''.join(tui_away_inning_runs)}"
    if last_chunk:
        linescore += "".join(tui_away_totals)
    linescore += f"\n{home_team_id}{''.join(tui_home_inning_runs)}"
    if last_chunk:
        linescore += "".join(tui_home_totals)
    return linescore


def get_tui_inning_number(inning_number):
    if isinstance(inning_number, str):
        return ""
    pad_len = 1 if inning_number < 10 else 0
    return f"{' '*pad_len}{inning_number}"


def add_filler_columns(list_to_pad):
    for _ in range(9 - len(list_to_pad)):
        list_to_pad.append("")
    return list_to_pad


def get_team_record_for_linescore(team_data):
    team_wins = team_data["total_wins_before_game"]
    team_losses = team_data["total_losses_before_game"]
    return f"{team_wins}-{team_losses}"


def get_team_totals_for_linescore(team_data):
    team_total_runs = team_data["total_runs_scored_by_team"]
    team_hits = team_data["total_hits_by_team"]
    team_errors = team_data["total_errors_by_team"]
    return [team_total_runs, team_hits, team_errors]


def convert_linescore_for_html(linescore):
    if isinstance(linescore, list):
        linescore = linescore[0]
    inning_runs_scored_map = get_inning_runs_scored_map(linescore)
    (_, removed_inning_map, _) = _condense_linescore(inning_runs_scored_map)
    html_dict_keys = ("col_index", "col_header", "away_team", "home_team", "css_class", "removed_inning")
    team_id_column = (0, "&nbsp;", linescore["away_team_id"], linescore["home_team_id"], "team-id", False)
    html_linescore = [dict(zip(html_dict_keys, team_id_column))]

    total_innings = len(linescore["inning_numbers"])
    inning_col_indices = list(range(1, total_innings + 1))
    inning_css_classes = ["inning-runs-scored" for _ in range(total_innings)]
    inning_columns = zip(
        inning_col_indices,
        linescore["inning_numbers"],
        linescore["away_team_runs_by_inning"],
        linescore["home_team_runs_by_inning"],
        inning_css_classes,
        removed_inning_map,
    )
    html_linescore.extend([dict(zip(html_dict_keys, column)) for column in inning_columns])

    game_total_col_indices = list(range(total_innings + 1, total_innings + 4))
    game_total_css_classes = ["game-total" for _ in range(len(linescore["game_totals"]))]
    removed_inning_false = [False for _ in range(len(linescore["game_totals"]))]
    game_total_columns = zip(
        game_total_col_indices,
        linescore["game_totals"],
        linescore["away_team_totals"],
        linescore["home_team_totals"],
        game_total_css_classes,
        removed_inning_false,
    )
    html_linescore.extend([dict(zip(html_dict_keys, column)) for column in game_total_columns])
    return html_linescore


def get_innings_sorted(innings_unsorted):
    inning_ids = {inning["inning_id"] for inning in innings_unsorted}
    inning_weights = {inning_id: get_inning_weight(inning_id) for inning_id in inning_ids}
    return sorted(innings_unsorted, key=lambda x: inning_weights[x["inning_id"]])


def get_inning_weight(inning_id):
    match = INNING_LABEL_REGEX.search(inning_id)
    if not match:
        return 0
    inn_half = match[1]
    inn_number = int(match[2])
    top_or_bottom = -1 if inn_half == "TOP" else 0
    return inn_number * 2 + top_or_bottom


def parse_inning_number_from_label(inning_id):
    match = INNING_LABEL_REGEX.search(inning_id)
    if not match:
        return 0
    try:
        return int(match[2])
    except ValueError:
        return 0


def get_at_bat_details(at_bat, pitch_app_stats, bat_stats):
    pitch_stats = parse_pitch_app_stats(pitch_app_stats)
    (at_bats, stat_line) = parse_bat_stats_for_game(bat_stats)
    details = f", {stat_line}" if stat_line else ""
    inning = inning_number_to_string(at_bat["inning_id"][-5:])
    score = parse_score_for_at_bat(at_bat)
    outs_plural = "out" if at_bat["outs_before_play"] == 1 else "outs"
    outs = f"{at_bat['outs_before_play']} {outs_plural}"
    at_bat_result = parse_at_bat_result(at_bat)
    return (
        f"Pitcher..: {at_bat['pitcher_name']} ({pitch_stats})\n"
        f"Batter...: {at_bat['batter_name']} ({at_bats}{details})\n"
        f"Score....: {score}\n"
        f"Inning...: {inning}, {outs}\n"
        f"On Base..: {at_bat['runners_on_base']}\n"
        f"Result...: {at_bat_result}"
    )


def parse_score_for_at_bat(at_bat):
    if not at_bat["pitchfx"]:
        return at_bat["score"]
    pfx = at_bat["pitchfx"][0]
    match = INNING_LABEL_REGEX.search(at_bat["inning_id"])
    if not match:
        return 0
    inn_half = match[1]
    batter_team_id = get_bbref_team_id(pfx["opponent_team_id_bb"])
    pitcher_team_id = get_bbref_team_id(pfx["pitcher_team_id_bb"])
    split = at_bat["score"].split("-")
    return (
        f"{batter_team_id} {split[0]} - {pitcher_team_id} {split[1]}"
        if inn_half == "TOP"
        else f"{pitcher_team_id} {split[1]} - {batter_team_id} {split[0]}"
    )


def parse_at_bat_result(at_bat):
    outcome = at_bat["play_description"]
    if at_bat["pitchfx"]:
        outcome = at_bat["pitchfx"][0]["des"]
    total_outs = at_bat["runs_outs_result"].count("O")
    total_runs = at_bat["runs_outs_result"].count("R")
    if not total_outs and not total_runs:
        return outcome
    outs_plural = "out" if total_outs == 1 else "outs"
    outs = f"{total_outs} {outs_plural}" if total_outs else ""
    runs_plural = "run" if total_runs == 1 else "runs"
    runs = f"{total_runs} {runs_plural} scored" if total_runs else ""
    separator = ", " if total_outs and total_runs else ""
    return f"{outcome} ({outs}{separator}{runs})"


def parse_bat_stats_for_game(bat_stats):
    if "bbref_data" not in bat_stats or not bat_stats["bbref_data"] or not bat_stats["bbref_data"]["plate_appearances"]:
        return ("0/0", "")
    at_bats = f"{bat_stats['bbref_data']['hits']}/{bat_stats['bbref_data']['at_bats']}"
    stat_line = get_detailed_bat_stats(bat_stats["bbref_data"])
    return (at_bats, stat_line)


def get_detailed_bat_stats(bat_stats):
    stats = homeruns = doubles = triples = stolen_bases = iw = None
    stat_list = []
    if bat_stats["details"]:
        stats = parse_bat_stat_details(bat_stats["details"])
        homeruns = stats.pop("HR", None)
        triples = stats.pop("3B", None)
        doubles = stats.pop("2B", None)
        stolen_bases = stats.pop("SB", None)
        iw = stats.pop("IW", None)
    if homeruns:
        stat_list.append(homeruns)
    if triples:
        stat_list.append(triples)
    if doubles:
        stat_list.append(doubles)
    if bat_stats["rbis"]:
        rbis = bat_stats["rbis"] if bat_stats["rbis"] > 1 else ""
        stat_list.append(f"{rbis}RBI")
    if bat_stats["runs_scored"]:
        runs_scored = bat_stats["runs_scored"] if bat_stats["runs_scored"] > 1 else ""
        stat_list.append(f"{runs_scored}R")
    if stolen_bases:
        stat_list.append(stolen_bases)
    if bat_stats["strikeouts"]:
        strikeouts = bat_stats["strikeouts"] if bat_stats["strikeouts"] > 1 else ""
        stat_list.append(f"{strikeouts}K")
    if bat_stats["bases_on_balls"]:
        bases_on_balls = bat_stats["bases_on_balls"] if bat_stats["bases_on_balls"] > 1 else ""
        bb = f"{bases_on_balls}BB"
        if iw:
            bb += f" ({iw})"
        stat_list.append(bb)
    if stats:
        stat_list.extend(list(stats.values()))
    return ", ".join(stat_list)


def parse_bat_stat_details(details):
    return {stat["stat"]: parse_single_bat_stat(stat) for stat in details}


def parse_single_bat_stat(stat):
    separator = "-" if stat["stat"] in ["2B", "3B"] else ""
    return f"{stat['count']}{separator}{stat['stat']}" if stat["count"] > 1 else stat["stat"]


def parse_bat_stats_to_date(bat_stats):
    return (
        (
            f"{format_stat_decimal(bat_stats['bbref_data']['avg_to_date'])}/"
            f"{format_stat_decimal(bat_stats['bbref_data']['obp_to_date'])}/"
            f"{format_stat_decimal(bat_stats['bbref_data']['slg_to_date'])}/"
            f"{format_stat_decimal(bat_stats['bbref_data']['ops_to_date'])}"
        )
        if "bbref_data" in bat_stats
        else ""
    )


def format_stat_decimal(stat_value):
    stat_with_leading_zero = f"{stat_value:0.3f}"
    return stat_with_leading_zero[1:]


def parse_pitch_app_stats(pitch_stats):
    if not pitch_stats or not pitch_stats["bbref_data"]:
        return ""
    bbref_stats = pitch_stats["bbref_data"]
    ip = bbref_stats["innings_pitched"]
    r = bbref_stats["runs"]
    er = bbref_stats["earned_runs"]
    runs = f"{er}ER" if er == r else f"{r}R, {er}ER"
    stat_list = []
    if bbref_stats["hits"]:
        h = bbref_stats["hits"] if bbref_stats["hits"] > 1 else ""
        stat_list.append(f"{h}H")
    if bbref_stats["strikeouts"]:
        k = bbref_stats["strikeouts"] if bbref_stats["strikeouts"] > 1 else ""
        stat_list.append(f"{k}K")
    if bbref_stats["bases_on_balls"]:
        bb = bbref_stats["bases_on_balls"] if bbref_stats["bases_on_balls"] > 1 else ""
        stat_list.append(f"{bb}BB")
    details = f", {', '.join(stat_list)}" if stat_list else ""
    game_score = f", GS: {bbref_stats['game_score']}" if bbref_stats["game_score"] > 0 else ""
    return f"{ip} IP ({runs}{details}{game_score})"


def format_pitch_sequence_description(pitch_seq_desc):
    return [format_event_description(seq) for seq in pitch_seq_desc]


def format_event_description(event_desc):
    if not event_desc[0] and not event_desc[2] and event_desc[1].startswith("(") and event_desc[1].endswith(")"):
        event_desc[1] = event_desc[1][1:]
        event_desc[1] = event_desc[1][:-1]
    desc_lines = [wrap_text(s, max_len=AT_BAT_DESC_MAX_WIDTH) for s in event_desc[1].split("\n")]
    event_desc[1] = "\n".join(desc_lines).replace("\n\n", "\n")
    return event_desc


def check_pitch_stats_for_game_score(pitch_stats):
    if not pitch_stats or not pitch_stats["bbref_data"]:
        return False
    return pitch_stats["bbref_data"]["game_score"] > 0



from dataclasses import dataclass


@dataclass
class BBRefPitchStats:
    """Statistics for a an individual pitching appearance in a single game."""

    player_id_br: str = "0"
    player_team_id_br: str = "0"
    opponent_team_id_br: str = "0"
    innings_pitched: str = "0"
    hits: str = "0"
    runs: str = "0"
    earned_runs: str = "0"
    bases_on_balls: str = "0"
    strikeouts: str = "0"
    homeruns: str = "0"
    batters_faced: str = "0"
    pitch_count: str = "0"
    strikes: str = "0"
    strikes_contact: str = "0"
    strikes_swinging: str = "0"
    strikes_looking: str = "0"
    ground_balls: str = "0"
    fly_balls: str = "0"
    line_drives: str = "0"
    unknown_type: str = "0"
    game_score: str = "0"
    inherited_runners: str = "0"
    inherited_scored: str = "0"
    wpa_pitch: str = "0"
    avg_lvg_index: str = "0"
    re24_pitch: str = "0"

    def as_dict(self):
        """Convert a single player's pitching appearance to a dictionary."""
        return {
            "player_id_br": f"{self.player_id_br}",
            "player_team_id_br": f"{self.player_team_id_br}",
            "opponent_team_id_br": f"{self.opponent_team_id_br}",
            "innings_pitched": float(self.innings_pitched),
            "hits": int(self.hits),
            "runs": int(self.runs),
            "earned_runs": int(self.earned_runs),
            "bases_on_balls": int(self.bases_on_balls),
            "strikeouts": int(self.strikeouts),
            "homeruns": int(self.homeruns),
            "batters_faced": int(self.batters_faced),
            "pitch_count": int(self.pitch_count),
            "strikes": int(self.strikes),
            "strikes_contact": int(self.strikes_contact),
            "strikes_swinging": int(self.strikes_swinging),
            "strikes_looking": int(self.strikes_looking),
            "ground_balls": int(self.ground_balls),
            "fly_balls": int(self.fly_balls),
            "line_drives": int(self.line_drives),
            "unknown_type": int(self.unknown_type),
            "game_score": int(self.game_score),
            "inherited_runners": int(self.inherited_runners),
            "inherited_scored": int(self.inherited_scored),
            "wpa_pitch": float(self.wpa_pitch),
            "avg_lvg_index": float(self.avg_lvg_index),
            "re24_pitch": float(self.re24_pitch),
        }



import vigorish.database as db
from vigorish.cli.components.viewers import DictListTableViewer, DisplayPage, PageViewer
from vigorish.enums import StatusReport
from vigorish.util.datetime_util import get_date_range
from vigorish.util.dt_format_strings import DATE_MONTH_NAME, DATE_ONLY
from vigorish.util.list_helpers import flatten_list2d, make_chunked_list
from vigorish.util.result import Result
from vigorish.util.string_helpers import wrap_text


def report_status_single_date(db_session, game_date, report_type):
    pages = []
    result = _validate_single_date(db_session, game_date)
    if result.failure:
        return result
    date_status = result.value
    date_str = game_date.strftime(DATE_MONTH_NAME)
    heading = f"### OVERALL STATUS FOR {date_str} ###"
    pages.append(DisplayPage(date_status.status_report(), heading))
    if report_type == StatusReport.SINGLE_DATE_WITH_GAME_STATUS:
        game_status_dict = date_status.games_status_report()
        for num, (game_id, game_report) in enumerate(game_status_dict.items(), start=1):
            heading = f"### STATUS FOR {game_id} (Game {num}/{len(game_status_dict)}) ###"
            pages.append(DisplayPage(game_report, heading))
    if report_type in [
        StatusReport.DATE_DETAIL_MISSING_PITCHFX,
        StatusReport.SINGLE_DATE_WITH_GAME_STATUS,
    ]:
        pages.extend(_get_missing_pfx_data_report_for_date(db_session, date_status))
    return Result.Ok(_create_report_viewer(pages, text_color="bright_magenta"))


def report_season_status(db_session, year, report_type):
    if report_type == StatusReport.NONE:
        return Result.Fail("no report")
    season = db.Season.find_by_year(db_session, year)
    if report_type == StatusReport.SEASON_SUMMARY:
        heading = f"### STATUS REPORT FOR {season.name} ###"
        pages = [DisplayPage(season.status_report(), heading)]
        return Result.Ok(_create_report_viewer(pages, text_color="bright_yellow"))
    return report_date_range_status(db_session, season.start_date, season.end_date, report_type)


def report_date_range_status(db_session, start_date, end_date, report_type):
    if report_type == StatusReport.NONE:
        return Result.Fail("no report")
    result = _construct_date_range_status(db_session, start_date, end_date, report_type)
    if result.failure:
        return result
    status_date_range = result.value
    return _get_report_for_date_range(db_session, start_date, end_date, status_date_range, report_type)


def _validate_single_date(db_session, game_date):
    season = db.Season.find_by_year(db_session, game_date.year)
    date_is_valid = db.Season.is_date_in_season(db_session, game_date).success
    date_str = game_date.strftime(DATE_ONLY)
    if not date_is_valid:
        error = (
            f"'{date_str}' is not within the {season.name}:\n"
            f"season_start_date: {season.start_date_str}\n"
            f"season_end_date: {season.end_date_str}"
        )
        return Result.Fail(error)
    date_status = db.DateScrapeStatus.find_by_date(db_session, game_date)
    if not date_status:
        error = f"scrape_status_date does not contain an entry for date: {date_str}"
        return Result.Fail(error)
    return Result.Ok(date_status)


def _construct_date_range_status(db_session, start_date, end_date, report_type):
    show_all = False
    if report_type in [
        StatusReport.DATE_SUMMARY_ALL_DATES,
        StatusReport.DATE_DETAIL_ALL_DATES,
        StatusReport.DATE_DETAIL_MISSING_PITCHFX,
    ]:
        show_all = True
    status_date_range = []
    for game_date in get_date_range(start_date, end_date):
        date_status = db.DateScrapeStatus.find_by_date(db_session, game_date)
        if not date_status:
            error = "scrape_status_date does not contain an entry for date: {game_date.strftime(DATE_ONLY)}"
            return Result.Fail(error)
        if not show_all and date_status.scraped_all_game_data:
            continue
        status_date_range.append(date_status)
    return Result.Ok(status_date_range)


def _get_report_for_date_range(db_session, start_date, end_date, status_date_range, report_type):
    if report_type in [
        StatusReport.DATE_DETAIL_MISSING_DATA,
        StatusReport.DATE_DETAIL_ALL_DATES,
    ]:
        return _get_detailed_report_for_date_range(db_session, status_date_range, False)
    if report_type == StatusReport.DATE_DETAIL_MISSING_PITCHFX:
        return _get_detailed_report_for_date_range(db_session, status_date_range, True)
    return _get_summary_report_for_date_range(start_date, end_date, status_date_range)


def _get_detailed_report_for_date_range(db_session, status_date_range, missing_pitchfx):
    pages = []
    for date_status in status_date_range:
        game_date_str = date_status.game_date.strftime(DATE_MONTH_NAME)
        heading = f"### STATUS REPORT FOR {game_date_str} ###"
        pages.append(DisplayPage(date_status.status_report(), heading))
        if missing_pitchfx:
            pages.extend(_get_missing_pfx_data_report_for_date(db_session, date_status))
    return Result.Ok(_create_report_viewer(pages, text_color="bright_cyan"))


def _get_missing_pfx_data_report_for_date(db_session, date_status):
    game_date_str = date_status.game_date.strftime(DATE_MONTH_NAME)
    heading = f"### MISSING PITCHFX DATA FOR {game_date_str} ###"
    missing_pfx_ids_list = _get_missing_pfx_ids_for_date(db_session, date_status)
    chunked_list = make_chunked_list(missing_pfx_ids_list, chunk_size=4)
    return [DisplayPage(chunk, heading) for chunk in chunked_list]


def _get_missing_pfx_ids_for_date(db_session, date_status):
    if date_status.scraped_all_pitchfx_logs:
        return ["All PitchFX logs have been scraped"]
    if not date_status.scraped_all_brooks_pitch_logs:
        return ["Pitch Appearance IDs without PitchFx data cannot be reported until all pitch logs have been scraped."]
    missing_ids = db.DateScrapeStatus.get_unscraped_pitch_app_ids_for_date(db_session, date_status.game_date)
    return flatten_list2d([_format_id_list(game_id, id_list) for game_id, id_list in missing_ids.items()])


def _format_id_list(game_id, pitch_app_id_list):
    missing_ids = f'MISSING PITCHFX DATA: {", ".join(pitch_app_id_list)}\n'
    return [f"GAME ID: {game_id}", wrap_text(missing_ids, max_len=70)]


def _get_summary_report_for_date_range(start_date, end_date, status_date_range):
    start_str = start_date.strftime(DATE_MONTH_NAME)
    end_str = end_date.strftime(DATE_MONTH_NAME)
    heading = f"### STATUS REPORT FOR {start_str} - {end_str} ###"
    if not status_date_range:
        pages = [DisplayPage(["All data has been scraped for all dates in the requested range"], heading)]
        return Result.Ok(_create_report_viewer(pages, text_color="bright_magenta"))
    dict_list = [{"game_date": ds.game_date_str, "status": ds.scrape_status_description} for ds in status_date_range]
    date_report = DictListTableViewer(
        dict_list,
        prompt="Press Enter to dismiss report",
        confirm_only=True,
        heading=heading,
        heading_color="bright_magenta",
        message=None,
        table_color="bright_magenta",
    )
    return Result.Ok(date_report)


def _create_report_viewer(pages, text_color):
    return PageViewer(
        pages,
        prompt="Press Enter to dismiss report",
        confirm_only=True,
        heading_color=text_color,
        text_color=text_color,
        wrap_text=False,
    )



# flake8: noqa
from vigorish.cli.components.viewers.csv_viewer import CsvFileTableViewer
from vigorish.cli.components.viewers.dict_viewer import DictListTableViewer
from vigorish.cli.components.viewers.display_page import DisplayPage
from vigorish.cli.components.viewers.display_table import DisplayTable
from vigorish.cli.components.viewers.page_viewer import PageViewer
from vigorish.cli.components.viewers.table_viewer import TableViewer
from vigorish.cli.components.viewers.util import create_display_table, create_table_viewer



from dataclasses import dataclass

from dataclass_csv import accept_whitespaces
from sqlalchemy import Column, Float, ForeignKey, Integer, String
from sqlalchemy.orm import relationship

import vigorish.database as db


class PitchStats(db.Base):

    __tablename__ = "pitch_stats"
    id = Column(Integer, primary_key=True)
    bbref_game_id = Column(String)
    player_id_mlb = Column(Integer)
    player_id_bbref = Column(String)
    player_team_id_bbref = Column(String)
    opponent_team_id_bbref = Column(String)
    is_sp = Column(Integer, default=0)
    is_rp = Column(Integer, default=0)
    is_wp = Column(Integer, default=0)
    is_lp = Column(Integer, default=0)
    is_sv = Column(Integer, default=0)
    innings_pitched = Column(Float)
    total_outs = Column(Integer)
    hits = Column(Integer)
    runs = Column(Integer)
    earned_runs = Column(Integer)
    bases_on_balls = Column(Integer)
    strikeouts = Column(Integer)
    homeruns = Column(Integer)
    batters_faced = Column(Integer)
    pitch_count = Column(Integer)
    strikes = Column(Integer)
    strikes_contact = Column(Integer)
    strikes_swinging = Column(Integer)
    strikes_looking = Column(Integer)
    ground_balls = Column(Integer)
    fly_balls = Column(Integer)
    line_drives = Column(Integer)
    unknown_type = Column(Integer)
    game_score = Column(Integer)
    inherited_runners = Column(Integer)
    inherited_scored = Column(Integer)
    wpa_pitch = Column(Float)
    avg_lvg_index = Column(Float)
    re24_pitch = Column(Float)
    pitch_app_id = Column(String)

    player_id = Column(Integer, ForeignKey("player.id"), index=True)
    player_team_id = Column(Integer, ForeignKey("team.id"))
    opponent_team_id = Column(Integer, ForeignKey("team.id"))
    pitch_app_db_id = Column(Integer, ForeignKey("scrape_status_pitch_app.id"), index=True)
    game_status_id = Column(Integer, ForeignKey("scrape_status_game.id"))
    date_id = Column(Integer, ForeignKey("scrape_status_date.id"))
    season_id = Column(Integer, ForeignKey("season.id"))

    player = relationship("Player", foreign_keys=[player_id], backref="pitch_stats")
    player_team = relationship("Team", foreign_keys=[player_team_id], backref="team_pitch_stats")
    opponent_team = relationship("Team", foreign_keys=[opponent_team_id], backref="opp_pitch_stats")
    game = relationship("GameScrapeStatus", foreign_keys=[game_status_id])
    date = relationship("DateScrapeStatus", foreign_keys=[date_id])
    season = relationship("Season", foreign_keys=[season_id])

    @classmethod
    def from_dict(cls, bbref_game_id, pitch_stats_dict):
        bbref_data = pitch_stats_dict.pop("bbref_data", {})
        bbref_data["bbref_game_id"] = bbref_game_id
        bbref_data["player_id_mlb"] = pitch_stats_dict["pitcher_id_mlb"]
        bbref_data["pitch_app_id"] = f'{bbref_game_id}_{pitch_stats_dict["pitcher_id_mlb"]}'
        bbref_data["player_id_bbref"] = pitch_stats_dict["pitcher_id_bbref"]
        bbref_data["player_team_id_bbref"] = pitch_stats_dict["pitcher_team_id_bbref"]
        bbref_data["opponent_team_id_bbref"] = pitch_stats_dict["opponent_team_id_bbref"]
        bbref_data["is_sp"] = int(pitch_stats_dict["is_sp"])
        bbref_data["is_rp"] = int(pitch_stats_dict["is_rp"])
        bbref_data["is_wp"] = int(pitch_stats_dict["is_wp"])
        bbref_data["is_lp"] = int(pitch_stats_dict["is_lp"])
        bbref_data["is_sv"] = int(pitch_stats_dict["is_sv"])
        bbref_data["total_outs"] = calculate_total_outs(bbref_data["innings_pitched"])
        return cls(**bbref_data)

    @classmethod
    def get_all_seasons_with_data_for_player(cls, db_session, player_id_mlb):
        return {ps.season for ps in db_session.query(cls).filter_by(player_id_mlb=player_id_mlb).all()}


def calculate_total_outs(innings_pitched):
    split = str(innings_pitched).split(".")
    if len(split) != 2:
        return 0
    inn_full = split[0]
    inn_partial = split[1]
    return int(inn_full) * 3 + int(inn_partial)


@accept_whitespaces
@dataclass
class PitchStatsCsvRow:
    id: int = None
    bbref_game_id: str = None
    player_id_mlb: int = None
    player_id_bbref: str = None
    player_team_id_bbref: str = None
    opponent_team_id_bbref: str = None
    is_sp: int = 0
    is_rp: int = 0
    is_wp: int = 0
    is_lp: int = 0
    is_sv: int = 0
    innings_pitched: float = 0.0
    total_outs: int = 0
    hits: int = 0
    runs: int = 0
    earned_runs: int = 0
    bases_on_balls: int = 0
    strikeouts: int = 0
    homeruns: int = 0
    batters_faced: int = 0
    pitch_count: int = 0
    strikes: int = 0
    strikes_contact: int = 0
    strikes_swinging: int = 0
    strikes_looking: int = 0
    ground_balls: int = 0
    fly_balls: int = 0
    line_drives: int = 0
    unknown_type: int = 0
    game_score: int = 0
    inherited_runners: int = 0
    inherited_scored: int = 0
    wpa_pitch: float = 0.0
    avg_lvg_index: float = 0.0
    re24_pitch: float = 0.0
    pitch_app_id: str = None



import re
from string import Template

from lxml import html
from w3lib.url import url_query_parameter

from vigorish.constants import DEFENSE_POSITIONS, VENUE_TERMS
from vigorish.scrape.bbref_boxscores.models.bat_stats import BBRefBatStats
from vigorish.scrape.bbref_boxscores.models.bat_stats_detail import BBRefBatStatsDetail
from vigorish.scrape.bbref_boxscores.models.boxscore import BBRefBoxscore
from vigorish.scrape.bbref_boxscores.models.boxscore_game_meta import BBRefBoxscoreMeta
from vigorish.scrape.bbref_boxscores.models.boxscore_team_data import BBRefBoxscoreTeamData
from vigorish.scrape.bbref_boxscores.models.half_inning import BBRefHalfInning
from vigorish.scrape.bbref_boxscores.models.pbp_event import BBRefPlayByPlayEvent
from vigorish.scrape.bbref_boxscores.models.pbp_other import BBRefPlayByPlayMiscEvent
from vigorish.scrape.bbref_boxscores.models.pbp_substitution import BBRefInGameSubstitution
from vigorish.scrape.bbref_boxscores.models.pitch_stats import BBRefPitchStats
from vigorish.scrape.bbref_boxscores.models.starting_lineup_slot import BBRefStartingLineupSlot
from vigorish.scrape.bbref_boxscores.models.umpire import BBRefUmpire
from vigorish.util.numeric_helpers import is_even
from vigorish.util.result import Result
from vigorish.util.string_helpers import fuzzy_match

_TEAM_ID_XPATH = '//a[@itemprop="name"]/@href'
_AWAY_TEAM_RECORD_XPATH = '//div[@class="scorebox"]/div[1]/div[3]/text()'
_HOME_TEAM_RECORD_XPATH = '//div[@class="scorebox"]/div[2]/div[3]/text()'
_SCOREBOX_META_XPATH = '//div[@class="scorebox_meta"]//div/text()'
_FIRST_PITCH_WEATHER_XPATH = '//strong[contains(text(), "Start Time Weather")]/../text()'
_LINESCORE_KEYS_XPATH = '//table[contains(@class, "linescore")]//th/text()'
_LINESCORE_AWAY_VALS_XPATH = '//table[contains(@class, "linescore")]//tbody/tr[1]//td/text()'
_LINESCORE_HOME_VALS_XPATH = '//table[contains(@class, "linescore")]//tbody/tr[2]//td/text()'
_LINESCORE_W_L_SV_XPATH = '//table[contains(@class, "linescore")]//tfoot/tr/td/text()'
_UMPIRES_XPATH = '//strong[contains(text(), "Umpires")]/../text()'

_BAT_STATS_TABLE_XPATH = '//table[contains(@id, "batting")][not(contains(@id, "clone"))]'
_BATTER_IDS_XPATH = './tbody//td[@data-stat="batting_avg"]/../th[@data-stat="player"]/@data-append-csv'
_BATTER_NAMES_XPATH = './tbody//td[@data-stat="batting_avg"]/../th[@data-stat="player"]//a/text()'
_T_BAT_STATS_ROW_XPATH = './tbody//th[@data-append-csv="${pid}"]/..'
_T_BAT_STATS_XPATH = './td[@data-stat="${stat}"]/text()'

_PITCH_STATS_TABLE_XPATH = '//table[contains(@id, "pitching")][not(contains(@id, "clone"))]'
_PITCHER_IDS_XPATH = './tbody//td[@data-stat="earned_run_avg"]/../th[@data-stat="player"]/@data-append-csv'
_PITCHER_NAMES_XPATH = './tbody//td[@data-stat="earned_run_avg"]/../th[@data-stat="player"]//a/text()'
_T_PITCH_STATS_ROW_XPATH = './tbody//th[@data-append-csv="${pid}"]/..'
_T_PITCH_STAT_XPATH = './td[@data-stat="${stat}"]/text()'

_AWAY_LINEUP_ORDER_XPATH = '//div[@id="lineups_1"]//table//tbody//tr//td[1]/text()'
_AWAY_LINEUP_PLAYER_XPATH = '//div[@id="lineups_1"]//table//tbody//a/@href'
_AWAY_LINEUP_DEF_POS_XPATH = '//div[@id="lineups_1"]//table//tbody//tr//td[3]/text()'
_HOME_LINEUP_ORDER_XPATH = '//div[@id="lineups_2"]//table//tbody//tr//td[1]/text()'
_HOME_LINEUP_PLAYER_XPATH = '//div[@id="lineups_2"]//table//tbody//a/@href'
_HOME_LINEUP_DEF_POS_XPATH = '//div[@id="lineups_2"]//table//tbody//tr//td[3]/text()'

_PBP_TABLE_XPATH = '//table[contains(@id, "play_by_play")][not(contains(@id, "clone"))]'
_PBP_INN_SUM_TOP_XPATH = './tbody//th[@data-stat="inning_summary_12"]/text()'
_PBP_INN_SUM_TOP_ROW_NUM_XPATH = './tbody//tr[@class="pbp_summary_top"]/@data-row'
_PBP_INN_SUM_BOT_XPATH = './tbody//tr[@class="pbp_summary_bottom"]//td[last()]/text()'
_PBP_INN_SUM_BOT_ROW_NUM_XPATH = './tbody//tr[@class="pbp_summary_bottom"]/@data-row'
_PBP_INN_SUM_BOT_LAST_XPATH = './tbody//tr[@class="pbp_summary_bottom"]//span[@class="half_inning_summary"]/text()'
_PBP_PLAYER_SUB_ROW_NUM_XPATH = './tbody//tr[@class="ingame_substitution"]/@data-row'
_T_PBP_PLAYER_SUB_XPATH = (
    './tbody//tr[@class="ingame_substitution"][@data-row="${row}"]' '//td[@data-stat="inning_summary_3"]//div/text()'
)
_T_PBP_MISC_EVENT_XPATH = './tbody//tr[@data-row="${row}"]//td[@data-stat="inning_summary_3"]//text()'
_PBP_INNING_XPATH = './tbody//tr[not(contains(@class, "thead"))]/th[@data-stat="inning"]/text()'

BAT_STATS = {
    "at_bats": "AB",
    "runs_scored": "R",
    "hits": "H",
    "rbis": "RBI",
    "bases_on_balls": "BB",
    "strikeouts": "SO",
    "plate_appearances": "PA",
    "avg_to_date": "batting_avg",
    "obp_to_date": "onbase_perc",
    "slg_to_date": "slugging_perc",
    "ops_to_date": "onbase_plus_slugging",
    "total_pitches": "pitches",
    "total_strikes": "strikes_total",
    "wpa_bat": "wpa_bat",
    "avg_lvg_index": "leverage_index_avg",
    "wpa_bat_pos": "wpa_bat_pos",
    "wpa_bat_neg": "wpa_bat_neg",
    "re24_bat": "re24_bat",
    "details": "details",
}

PITCH_STATS = {
    "innings_pitched": "IP",
    "hits": "H",
    "runs": "R",
    "earned_runs": "ER",
    "bases_on_balls": "BB",
    "strikeouts": "SO",
    "homeruns": "HR",
    "batters_faced": "batters_faced",
    "pitch_count": "pitches",
    "strikes": "strikes_total",
    "strikes_contact": "strikes_contact",
    "strikes_swinging": "strikes_swinging",
    "strikes_looking": "strikes_looking",
    "ground_balls": "inplay_gb_total",
    "fly_balls": "inplay_fb_total",
    "line_drives": "inplay_ld",
    "unknown_type": "inplay_unk",
    "game_score": "game_score",
    "inherited_runners": "inherited_runners",
    "inherited_scored": "inherited_score",
    "wpa_pitch": "wpa_def",
    "avg_lvg_index": "leverage_index_avg",
    "re24_pitch": "re24_def",
}

PBP_STATS = {
    "pbp_table_row_number": "data-row",
    "score": "score_batting_team",
    "outs_before_play": "outs",
    "runners_on_base": "runners_on_bases_pbp",
    "team_batting_id_br": "batting_team_id",
    "play_description": "play_desc",
    "pitch_sequence": "pitches_pbp",
    "runs_outs_result": "runs_outs_result",
    "play_index_url": "play_index_url",
}

_GAME_ID_REGEX = re.compile(r"[A-Z]{3,3}\d{9,9}")
_TEAM_ID_REGEX = re.compile(r"[A-Z]{3,3}")
_W_L_SV_NAME_REGEX = re.compile(r"[LPSVW]{2,2}: (.+) (?:\(\d{1,2}-?\d{0,2}\))")
_ATTENDANCE_REGEX = re.compile(r"\d{1,2},\d{3,3}")
_GAME_DURATION_REGEX = re.compile(r"[1-9]:[0-5][0-9]")
_CHANGE_POS_REGEX = re.compile(r"from\s\b(?P<old_pos>\w+)\b\sto\s\b(?P<new_pos>\w+)\b")
_POS_REGEX = re.compile(r"\([BCDFHLPRS123]{1,2}\)")
_NUM_REGEX = re.compile(r"[1-9]{1}")
_INNING_TOTALS_REGEX = re.compile(
    r"(?P<runs>\d{1,2}) (run|runs), "
    r"(?P<hits>\d{1,2}) (hit|hits), "
    r"(?P<errors>\d{1,2}) (error|errors), "
    r"(?P<left_on_base>-?\d{1,2}) LOB."
    r"(?P<away_team_name>\s\b\w+\b){1,2} "
    r"(?P<away_team_runs>\d{1,2}),"
    r"(?P<home_team_name>\s\b\w+\b){1,2} "
    r"(?P<home_team_runs>\d{1,2})"
)


def parse_bbref_boxscore(scraped_html, url, game_id):
    """Parse boxscore data from the page source."""
    page_content = html.fromstring(scraped_html, base_url=url)
    (away_team_id, home_team_id) = _parse_team_ids(page_content, url)

    (
        away_team_bat_table,
        home_team_bat_table,
        away_team_pitch_table,
        home_team_pitch_table,
        play_by_play_table,
    ) = _parse_data_tables(page_content)

    (away_team_data, home_team_data) = _parse_all_team_data(
        page_content,
        away_team_bat_table,
        home_team_bat_table,
        away_team_pitch_table,
        home_team_pitch_table,
        away_team_id,
        home_team_id,
    )

    player_name_dict = _create_player_name_dict(
        away_team_bat_table,
        home_team_bat_table,
        away_team_pitch_table,
        home_team_pitch_table,
        away_team_id,
        home_team_id,
    )

    result = _parse_w_l_sv_pitchers(page_content, away_team_data, home_team_data, player_name_dict)
    if result.failure:
        return result

    result = _parse_game_meta_info(page_content)
    if result.failure:
        return result
    game_meta_info = result.value
    umpires = _parse_umpires(page_content)
    if not umpires:
        error = "Failed to parse umpire info"
        return Result.Fail(error)
    result = _parse_all_game_events(play_by_play_table, game_id, away_team_id, home_team_id, player_name_dict)
    if result.failure:
        error = f"Failed to parse innnings list:\n{result.error}"
        return Result.Fail(error)
    result_dict = result.value
    innings_list = result_dict["innings_list"]
    player_id_match_log = result_dict["player_id_match_log"]
    player_team_dict = result_dict["player_team_dict"]
    player_name_dict = {f"{name}, {team_id}": player_id for (name, team_id), player_id in player_name_dict.items()}

    boxscore = BBRefBoxscore(
        boxscore_url=url,
        bbref_game_id=game_id,
        game_meta_info=game_meta_info,
        away_team_data=away_team_data,
        home_team_data=home_team_data,
        innings_list=innings_list,
        umpires=umpires,
        player_id_match_log=player_id_match_log,
        player_team_dict=player_team_dict,
        player_name_dict=player_name_dict,
    )
    return Result.Ok(boxscore)


def _parse_team_ids(page_content, url):
    away_team_id = _parse_away_team_id(page_content)
    if not away_team_id:
        error = "Failed to parse away team ID"
        return Result.Fail(error)
    home_team_id = _parse_home_team_id(page_content)
    if not home_team_id:
        error = "Failed to parse home team ID"
        return Result.Fail(error)
    return (away_team_id, home_team_id)


def _parse_data_tables(page_content):
    result = page_content.xpath(_BAT_STATS_TABLE_XPATH)
    if not result or len(result) != 2:
        error = "Failed to parse batting stats table"
        return Result.Fail(error)
    away_team_bat_table = result[0]
    home_team_bat_table = result[1]

    result = page_content.xpath(_PITCH_STATS_TABLE_XPATH)
    if not result or len(result) != 2:
        error = "Failed to parse pitching stats table"
        return Result.Fail(error)
    away_team_pitch_table = result[0]
    home_team_pitch_table = result[1]

    result = page_content.xpath(_PBP_TABLE_XPATH)
    if not result:
        error = "Failed to parse play by play table"
        return Result.Fail(error)
    play_by_play_table = result[0]

    return (
        away_team_bat_table,
        home_team_bat_table,
        away_team_pitch_table,
        home_team_pitch_table,
        play_by_play_table,
    )


def _parse_away_team_id(page_content):
    name_urls = page_content.xpath(_TEAM_ID_XPATH)
    if not name_urls:
        return None
    matches = _TEAM_ID_REGEX.findall(name_urls[0])
    return matches[0] if matches else None


def _parse_home_team_id(page_content):
    name_urls = page_content.xpath(_TEAM_ID_XPATH)
    if not name_urls:
        return None
    matches = _TEAM_ID_REGEX.findall(name_urls[1])
    return matches[0] if matches else None


def _parse_all_team_data(
    page_content,
    away_team_bat_table,
    home_team_bat_table,
    away_team_pitch_table,
    home_team_pitch_table,
    away_team_id,
    home_team_id,
):
    result = _parse_team_data(
        page_content,
        team_batting_table=away_team_bat_table,
        team_pitching_table=away_team_pitch_table,
        team_id=away_team_id,
        opponent_id=home_team_id,
        is_home_team=False,
    )
    if result.failure:
        return result
    away_team_dict = result.value
    away_team_data = BBRefBoxscoreTeamData(**away_team_dict)

    result = _parse_team_data(
        page_content,
        team_batting_table=home_team_bat_table,
        team_pitching_table=home_team_pitch_table,
        team_id=home_team_id,
        opponent_id=away_team_id,
        is_home_team=True,
    )
    if result.failure:
        return result
    home_team_dict = result.value
    home_team_data = BBRefBoxscoreTeamData(**home_team_dict)

    return (away_team_data, home_team_data)


def _parse_team_data(page_content, team_batting_table, team_pitching_table, team_id, opponent_id, is_home_team):
    team_record = _parse_team_record(page_content, is_home_team)
    if not team_record:
        error = f"Failed to parse team record (team_id={team_id}, is_home_team={is_home_team})"
        return Result.Fail(error)
    team_totals = _parse_game_totals_for_team(page_content, team_id, is_home_team)
    if not team_totals:
        error = "Failed to parse team linescore totals (team_id={team_id}, is_home_team={is_home_team})"
        return Result.Fail(error)
    opponent_totals = _parse_game_totals_for_team(page_content, opponent_id, not is_home_team)
    if not opponent_totals:
        error = "Failed to parse opponent linescore totals (team_id={opponent_id}, is_home_team={not is_home_team})"
        return Result.Fail(error)
    batting_stats = _parse_batting_stats_for_team(team_batting_table, team_id, opponent_id)
    if not batting_stats:
        error = "Failed to parse away team batting stats"
        return Result.Fail(error)
    pitching_stats = _parse_pitching_stats_for_team(team_pitching_table, team_id, opponent_id)
    if not pitching_stats:
        error = "Failed to parse away team pitching stats"
        return Result.Fail(error)
    result = _parse_starting_lineup_for_team(page_content, team_id, is_home_team)
    if result.failure:
        return result
    starting_lineup = result.value

    team_dict = {
        "team_id_br": team_id,
        "total_wins_before_game": int(team_record[0]),
        "total_losses_before_game": int(team_record[1]),
        "total_runs_scored_by_team": int(team_totals["R"]),
        "total_runs_scored_by_opponent": int(opponent_totals["R"]),
        "total_hits_by_team": int(team_totals["H"]),
        "total_hits_by_opponent": int(opponent_totals["H"]),
        "total_errors_by_team": int(team_totals["E"]),
        "total_errors_by_opponent": int(opponent_totals["E"]),
        "batting_stats": batting_stats,
        "pitching_stats": pitching_stats,
        "starting_lineup": starting_lineup,
    }
    return Result.Ok(team_dict)


def _parse_team_record(page_content, is_home_team):
    if is_home_team:
        team_record_xpath = _HOME_TEAM_RECORD_XPATH
    else:
        team_record_xpath = _AWAY_TEAM_RECORD_XPATH
    team_record_dirty = page_content.xpath(team_record_xpath)
    if not team_record_dirty:
        return None
    team_record = team_record_dirty[0].split("-")
    if len(team_record) != 2:
        return None
    return team_record


def _parse_game_totals_for_team(page_content, team_id, is_home_team):
    if is_home_team:
        linescore_xpath = _LINESCORE_HOME_VALS_XPATH
    else:
        linescore_xpath = _LINESCORE_AWAY_VALS_XPATH

    keys = page_content.xpath(_LINESCORE_KEYS_XPATH)
    vals = page_content.xpath(linescore_xpath)
    if not keys or not vals:
        return None
    if len(keys) < 2 or len(vals) < 2:
        return None
    del keys[0:2]
    del vals[0:2]
    if len(keys) < 3 or len(vals) < 3:
        return None
    del keys[:-3]
    del vals[:-3]
    if len(keys) != len(vals):
        return None
    return dict(zip(keys, vals))


def _parse_batting_stats_for_team(team_batting_table, player_team_id, opponent_team_id):
    bat_data = []
    for player_id in team_batting_table.xpath(_BATTER_IDS_XPATH):
        player_bat_stats_xpath = Template(_T_BAT_STATS_ROW_XPATH).substitute(pid=player_id)
        results = team_batting_table.xpath(player_bat_stats_xpath)
        if not results:
            return []
        stat_dict = {attr: _get_bat_stat_value(results[0], stat_name) for attr, stat_name in BAT_STATS.items()}
        stat_dict["player_id_br"] = player_id
        stat_dict["player_team_id_br"] = player_team_id
        stat_dict["opponent_team_id_br"] = opponent_team_id
        stat_dict["details"] = _parse_batting_details(stat_dict)
        player_data = BBRefBatStats(**stat_dict)
        bat_data.append(player_data)
    return bat_data


def _get_bat_stat_value(player_bat_data_html, stat_name):
    bat_stat_xpath = Template(_T_BAT_STATS_XPATH).substitute(stat=stat_name)
    bat_stat_value = "0"
    if stat_name == "details":
        bat_stat_value = []
    result = player_bat_data_html.xpath(bat_stat_xpath)
    if result and len(result) > 0:
        bat_stat_value = result[0]
    return bat_stat_value if isinstance(bat_stat_value, list) else bat_stat_value.rstrip("%")


def _parse_batting_details(stat_dict):
    details_str = stat_dict["details"]
    if not details_str:
        return []
    details = []
    for s in details_str.split(","):
        d = BBRefBatStatsDetail()
        t = s.split("\u00b7")
        if len(t) == 1:
            d.count = "1"
            d.stat = t[0].strip("\n")
            details.append(d)
        if len(t) == 2:
            d.count = t[0][0]
            d.stat = t[1].strip("\n")
            details.append(d)
    return details


def _parse_pitching_stats_for_team(team_pitching_table, player_team_id, opponent_team_id):
    pitch_data = []
    for player_id in team_pitching_table.xpath(_PITCHER_IDS_XPATH):
        player_pitch_stats_xpath = Template(_T_PITCH_STATS_ROW_XPATH).substitute(pid=player_id)
        results = team_pitching_table.xpath(player_pitch_stats_xpath)
        if not results:
            return None
        stat_dict = {attr: _get_pitch_stat_value(results[0], stat_name) for attr, stat_name in PITCH_STATS.items()}
        stat_dict["player_id_br"] = player_id
        stat_dict["player_team_id_br"] = player_team_id
        stat_dict["opponent_team_id_br"] = opponent_team_id
        player_data = BBRefPitchStats(**stat_dict)
        pitch_data.append(player_data)
    return pitch_data


def _get_pitch_stat_value(player_pitch_data_html, stat_name):
    pitch_stat_xpath = Template(_T_PITCH_STAT_XPATH).substitute(stat=stat_name)
    pitch_stat_value = "0"
    result = player_pitch_data_html.xpath(pitch_stat_xpath)
    if result and len(result) > 0:
        pitch_stat_value = result[0]
    return pitch_stat_value if isinstance(pitch_stat_value, list) else pitch_stat_value.rstrip("%")


def _parse_starting_lineup_for_team(page_content, team_id, is_home_team):
    lineup = []
    if is_home_team:
        bat_orders = page_content.xpath(_HOME_LINEUP_ORDER_XPATH)
        player_links = page_content.xpath(_HOME_LINEUP_PLAYER_XPATH)
        def_positions = page_content.xpath(_HOME_LINEUP_DEF_POS_XPATH)
    else:
        bat_orders = page_content.xpath(_AWAY_LINEUP_ORDER_XPATH)
        player_links = page_content.xpath(_AWAY_LINEUP_PLAYER_XPATH)
        def_positions = page_content.xpath(_AWAY_LINEUP_DEF_POS_XPATH)
    if not bat_orders or not player_links or not def_positions:
        error = "Failed to parse team lineup data (team_id={team_id}, is_home_team={is_home_team})"
        return Result.Fail(error)
    for i in range(len(bat_orders)):
        bat = BBRefStartingLineupSlot()
        split = player_links[i].split("/")
        if len(split) != 4:
            error = (
                "Player links in lineup table are not in the expected format "
                f"(team_id={team_id}, is_home_team={is_home_team})"
            )
            return Result.Fail(error)
        bat.player_id_br = split[3][:-6]
        bat.bat_order = bat_orders[i]
        bat.def_position = def_positions[i]
        lineup.append(bat)
    return Result.Ok(lineup)


def _parse_w_l_sv_pitchers(page_content, away_team_data, home_team_data, player_name_dict):
    pitcher_names = _parse_w_l_sv_save_pitcher_names(page_content)
    if not pitcher_names["wp"] or not pitcher_names["lp"]:
        return Result.Fail("Failed to parse the names of the WP and LP")
    winning_team_id = away_team_data.team_id_br if away_team_data.team_won else home_team_data.team_id_br
    losing_team_id = away_team_data.team_id_br if home_team_data.team_won else home_team_data.team_id_br
    w_pitcher_id_br = _match_player_id(pitcher_names["wp"], winning_team_id, player_name_dict)["id"]
    l_pitcher_id_br = _match_player_id(pitcher_names["lp"], losing_team_id, player_name_dict)["id"]
    sv_pitcher_id_br = (
        _match_player_id(pitcher_names["sv"], winning_team_id, player_name_dict)["id"] if pitcher_names["sv"] else ""
    )
    if away_team_data.team_won:
        away_team_data.pitcher_of_record = w_pitcher_id_br
        away_team_data.pitcher_earned_save = sv_pitcher_id_br
        home_team_data.pitcher_of_record = l_pitcher_id_br
    else:
        away_team_data.pitcher_of_record = l_pitcher_id_br
        home_team_data.pitcher_of_record = w_pitcher_id_br
        home_team_data.pitcher_earned_save = sv_pitcher_id_br
    return Result.Ok()


def _parse_w_l_sv_save_pitcher_names(page_content):
    pitchers = page_content.xpath(_LINESCORE_W_L_SV_XPATH)
    if not pitchers:
        return None
    pitchers = pitchers[0].strip().replace("\xa0", " ")
    pitcher_list = pitchers.split(" • ")
    if len(pitcher_list) not in [2, 3]:
        return None
    wp_match = _W_L_SV_NAME_REGEX.search(pitcher_list[0])
    lp_match = _W_L_SV_NAME_REGEX.search(pitcher_list[1])
    sv_match = _W_L_SV_NAME_REGEX.search(pitcher_list[2]) if len(pitcher_list) == 3 else None
    return {
        "wp": wp_match.group(1) if wp_match else "",
        "lp": lp_match.group(1) if lp_match else "",
        "sv": sv_match.group(1) if sv_match else "",
    }


def _parse_game_meta_info(page_content):
    scorebox_meta_strings = page_content.xpath(_SCOREBOX_META_XPATH)
    attendance_matches = _parse_attendance_from_strings(scorebox_meta_strings)
    if attendance_matches is not None:
        attendance = attendance_matches["match"]
        attendance_index = attendance_matches["index"]
        del scorebox_meta_strings[attendance_index]
    else:
        attendance = "0"

    venue_matches = _parse_venue_from_strings(scorebox_meta_strings)
    if venue_matches is None:
        error = "Failed to parse park name"
        return Result.Fail(error)

    park_name = venue_matches["match"]
    venue_index = venue_matches["index"]
    del scorebox_meta_strings[venue_index]

    game_duration_matches = _parse_game_duration_from_strings(scorebox_meta_strings)
    if game_duration_matches is None:
        error = "Failed to parse game duration"
        return Result.Fail(error)

    game_duration = game_duration_matches["match"]
    game_duration_index = game_duration_matches["index"]
    del scorebox_meta_strings[game_duration_index]

    day_night_field = _parse_day_night_field_type_from_strings(scorebox_meta_strings)
    if day_night_field is None:
        error = "Failed to parse game time/field type"
        return Result.Fail(error)

    split = day_night_field["match"].split(",")
    day_night_field_index = day_night_field["index"]
    del scorebox_meta_strings[day_night_field_index]

    if len(split) < 2:
        error = "Failed to parse game time/field type"
        return Result.Fail(error)

    day_night = split[0].strip()
    field_type = split[1].strip().title()

    first_pitch_weather = page_content.xpath(_FIRST_PITCH_WEATHER_XPATH)
    if not first_pitch_weather:
        error = "Failed to parse first pitch weather info"
        return Result.Fail(error)
    split2 = first_pitch_weather[0].split(",")
    if len(split2) < 2:
        error = f"First pitch weather info is not in expected format:\n{first_pitch_weather}"
        return Result.Fail(error)

    first_pitch_temperature = split2[0].strip()[:2]
    first_pitch_wind = split2[1].strip()

    first_pitch_clouds = ""
    if len(split2) > 2:
        first_pitch_clouds = split2[2].strip().strip(".")

    first_pitch_precipitation = ""
    if len(split2) > 3:
        first_pitch_precipitation = split2[3].strip().strip(".")

    game_info = BBRefBoxscoreMeta(
        attendance=int(attendance),
        park_name=park_name,
        game_duration=game_duration,
        day_night=day_night,
        field_type=field_type,
        first_pitch_temperature=int(first_pitch_temperature),
        first_pitch_wind=first_pitch_wind,
        first_pitch_clouds=first_pitch_clouds,
        first_pitch_precipitation=first_pitch_precipitation,
    )
    return Result.Ok(game_info)


def _parse_attendance_from_strings(strings):
    attendance = []
    for i in range(len(strings)):
        matches = _ATTENDANCE_REGEX.findall(strings[i])
        if matches:
            for m in matches:
                d = {"match": m.replace(",", ""), "index": i}
                attendance.append(d)
    if len(attendance) == 1:
        return attendance[0]
    return None


def _parse_venue_from_strings(strings):
    for i in range(len(strings)):
        for t in VENUE_TERMS:
            if t in strings[i].lower():
                return {"match": strings[i][2:], "index": i}
    return None


def _parse_game_duration_from_strings(strings):
    duration = []
    for i in range(len(strings)):
        if "start time" in strings[i].lower():
            continue
        matches = _GAME_DURATION_REGEX.findall(strings[i])
        if matches:
            for m in matches:
                d = {"match": m, "index": i}
                duration.append(d)
    if len(duration) == 1:
        return duration[0]
    return None


def _parse_day_night_field_type_from_strings(strings):
    matches = []
    for i in range(len(strings)):
        if "game, on" in strings[i].lower():
            d = {"match": strings[i], "index": i}
            matches.append(d)
    if len(matches) == 1:
        return matches[0]
    return None


def _parse_umpires(page_content):
    umpires = page_content.xpath(_UMPIRES_XPATH)[0]
    if umpires is None:
        return None
    split = umpires.split(",")
    umps = []
    for s in split:
        u = BBRefUmpire()
        split2 = s.strip(".").strip().split("-")
        if len(split2) != 2:
            return None
        u.field_location = split2[0].strip()
        u.umpire_name = split2[1].strip()
        umps.append(u)
    return umps


def _create_player_name_dict(
    away_team_bat_table,
    home_team_bat_table,
    away_team_pitch_table,
    home_team_pitch_table,
    away_team_id,
    home_team_id,
):
    away_team_batter_name_dict = _parse_batter_name_dict(away_team_bat_table, away_team_id)
    if not away_team_batter_name_dict:
        error = "Failed to parse away team batter name dictionary"
        return Result.Fail(error)
    home_team_batter_name_dict = _parse_batter_name_dict(home_team_bat_table, home_team_id)
    if not home_team_batter_name_dict:
        error = "Failed to parse home team batter name dictionary"
        return Result.Fail(error)
    away_team_pitcher_name_dict = _parse_pitcher_name_dict(away_team_pitch_table, away_team_id)
    if not away_team_pitcher_name_dict:
        error = "Failed to parse away team pitcher name dictionary"
        return Result.Fail(error)
    home_team_pitcher_name_dict = _parse_pitcher_name_dict(home_team_pitch_table, home_team_id)
    if not home_team_pitcher_name_dict:
        error = "Failed to parse home team pitcher name dictionary"
        return Result.Fail(error)
    batter_name_dict = {**away_team_batter_name_dict, **home_team_batter_name_dict}
    pitcher_name_dict = {**away_team_pitcher_name_dict, **home_team_pitcher_name_dict}
    return {**batter_name_dict, **pitcher_name_dict}


def _parse_batter_name_dict(team_batting_table, team_id_br):
    batter_ids = team_batting_table.xpath(_BATTER_IDS_XPATH)
    batter_names = team_batting_table.xpath(_BATTER_NAMES_XPATH)
    if not batter_ids or not batter_names:
        return None
    if len(batter_ids) != len(batter_names):
        return None
    dict_keys = [(name, team_id_br) for name in batter_names]
    return dict(zip(dict_keys, batter_ids))


def _parse_pitcher_name_dict(team_pitching_table, team_id_br):
    pitcher_ids = team_pitching_table.xpath(_PITCHER_IDS_XPATH)
    pitcher_names = team_pitching_table.xpath(_PITCHER_NAMES_XPATH)
    if not pitcher_ids or not pitcher_names:
        return None
    if len(pitcher_ids) != len(pitcher_names):
        return None
    dict_keys = [(name, team_id_br) for name in pitcher_names]
    return dict(zip(dict_keys, pitcher_ids))


def _parse_all_game_events(play_by_play_table, game_id, away_team_id, home_team_id, player_name_dict):
    inning_summaries_top = _parse_inning_summary_top(play_by_play_table)
    if not inning_summaries_top:
        error = "Failed to parse inning start summaries"
        return Result.Fail(error)

    inning_summaries_bottom = _parse_inning_summary_bottom(play_by_play_table)
    if not inning_summaries_bottom:
        error = "Failed to parse inning end summaries"
        return Result.Fail(error)

    result = _parse_play_by_play(play_by_play_table, player_name_dict, away_team_id, home_team_id, game_id)
    if result.failure:
        error = f"rescrape_did_not_parse_play_by_play:\n{result.error}"
        return Result.Fail(error)
    result_dict = result.value
    play_by_play_events = result_dict["play_by_play"]
    player_id_match_log = list(result_dict["player_id_match_log"])
    player_team_dict = _create_player_team_dict(play_by_play_events)
    if not player_team_dict:
        error = "Player name was unmatched in play by play events"
        return Result.Fail(error)

    result = _parse_in_game_substitutions(play_by_play_table, game_id, player_name_dict)
    if result.failure:
        error = f"Failed to parse in game substitutions:\n{result.error}"
        return Result.Fail(error)
    result_dict = result.value
    in_game_substitutions = result_dict["sub_list"]

    misc_events = []
    missing_row_ids = _find_missing_pbp_events(
        inning_summaries_top,
        inning_summaries_bottom,
        play_by_play_events,
        in_game_substitutions,
    )
    if missing_row_ids:
        misc_events = _parse_missing_pbp_events(missing_row_ids, play_by_play_table, game_id)

    result = _create_innings_list(
        game_id,
        away_team_id,
        home_team_id,
        player_name_dict,
        player_id_match_log,
        inning_summaries_top,
        inning_summaries_bottom,
        play_by_play_events,
        in_game_substitutions,
        misc_events,
    )
    if result.failure:
        error = f"Error occurred constructing inning list:\n{result.error}"
        return Result.Fail(error)
    (innings_list, player_id_match_log) = result.value
    result_dict = {
        "innings_list": innings_list,
        "player_id_match_log": player_id_match_log,
        "player_team_dict": player_team_dict,
    }
    return Result.Ok(result_dict)


def _parse_inning_summary_top(play_by_play_table):
    summaries = play_by_play_table.xpath(_PBP_INN_SUM_TOP_XPATH)
    if summaries is None:
        return None
    inning_summaries = [s.strip().replace("\xa0", " ") for s in summaries]
    row_numbers = play_by_play_table.xpath(_PBP_INN_SUM_TOP_ROW_NUM_XPATH)
    if row_numbers is None:
        return None
    summary_row_numbers = [int(s) for s in row_numbers]
    if len(inning_summaries) != len(summary_row_numbers):
        return None
    return dict(zip(summary_row_numbers, inning_summaries))


def _parse_inning_summary_bottom(play_by_play_table):
    summaries = play_by_play_table.xpath(_PBP_INN_SUM_BOT_XPATH)
    if summaries is None:
        return None
    inning_summaries = [s.strip() for s in summaries]
    last_summary = play_by_play_table.xpath(_PBP_INN_SUM_BOT_LAST_XPATH)
    for s in last_summary:
        inning_summaries.append(s.strip())
    row_numbers = play_by_play_table.xpath(_PBP_INN_SUM_BOT_ROW_NUM_XPATH)
    if row_numbers is None:
        return None
    summary_row_numbers = [int(s) for s in row_numbers]
    if len(inning_summaries) != len(summary_row_numbers):
        return None
    return dict(zip(summary_row_numbers, inning_summaries))


def _parse_play_by_play(pbp_table, player_id_dict, away_team_id, home_team_id, game_id):
    play_by_play = []
    player_id_match_log = []
    inning_list = pbp_table.xpath(_PBP_INNING_XPATH)
    for i in range(len(inning_list)):
        event_num = i + 1
        event_dict = _parse_pbp_event(pbp_table, event_num)
        inning_label = inning_list[i]
        inning_number = int(inning_label[-1])
        inning_id = f"{game_id}_INN_TOP{inning_number:02d}"
        event_dict["inning_label"] = inning_label
        event_dict["inning_id"] = inning_id

        if event_dict["team_batting_id_br"] == away_team_id:
            event_dict["team_pitching_id_br"] = home_team_id
        else:
            event_dict["team_pitching_id_br"] = away_team_id

        batter = _get_pbp_event_stat_value(pbp_table, "batter", event_num).replace("\xa0", " ")
        match = _match_player_id(batter, event_dict["team_batting_id_br"], player_id_dict)
        if match["type"] != "Exact match":
            player_id_match_log.append(match)
        event_dict["batter_id_br"] = match["id"]

        pitcher = _get_pbp_event_stat_value(pbp_table, "pitcher", event_num).replace("\xa0", " ")
        match = _match_player_id(pitcher, event_dict["team_pitching_id_br"], player_id_dict)
        if match["type"] != "Exact match":
            player_id_match_log.append(match)
        event_dict["pitcher_id_br"] = match["id"]

        if event_dict["play_index_url"]:
            event_dict["play_index_url"] = "https://www.baseball-reference.com" + event_dict["play_index_url"]
            event_dict["event_id"] = url_query_parameter(event_dict["play_index_url"], "game-event")

        event = BBRefPlayByPlayEvent(**event_dict)
        play_by_play.append(event)

    result = {"play_by_play": play_by_play, "player_id_match_log": player_id_match_log}
    return Result.Ok(result)


def _parse_pbp_event(pbp_table, event_number):
    return {
        attr: _get_pbp_event_stat_value(pbp_table, stat_name, event_number) for attr, stat_name in PBP_STATS.items()
    }


def _get_pbp_event_stat_value(pbp_table, stat_name, event_number):
    templ_xpath = './tbody//tr[@id="event_${n}"]/td[@data-stat="${s}"]/text()'
    query = Template(templ_xpath).substitute(n=event_number, s=stat_name)

    if stat_name == "pitches_pbp":
        templ_xpath = (
            './tbody//tr[@id="event_${n}"]/td[@data-stat="pitches_pbp"]' '/span[@class="pitch_sequence"]/text()'
        )
        query = Template(templ_xpath).substitute(n=event_number)

    if stat_name == "data-row":
        templ_xpath = './tbody//tr[@id="event_${n}"]/@data-row'
        query = Template(templ_xpath).substitute(n=event_number)

    if stat_name == "play_index_url":
        templ_xpath = './tbody//tr[@id="event_${n}"]/td[@data-stat="runners_on_bases_pbp"]/@data-endpoint'
        query = Template(templ_xpath).substitute(n=event_number)

    pbp_value = ""
    result = pbp_table.xpath(query)
    if result and len(result) > 0:
        pbp_value = result[0]
    return pbp_value


def _match_player_id(name, team_id, id_dict):
    match = {}
    if (name, team_id) in id_dict:
        match["type"] = "Exact match"
        match["name"] = name
        match["team_id"] = team_id
        match["id"] = id_dict[(name, team_id)]
        match["score"] = 1
    else:
        mapped_choices = {player_id: player_name for (player_name, team_id), player_id in id_dict.items()}
        best_matches = fuzzy_match(name, mapped_choices)
        player_id = best_matches[0]["result"]
        name_dict = {v: k for k, v in id_dict.items()}
        match["type"] = "Fuzzy match"
        match["name"] = name
        match["team_id"] = team_id
        match["best_match"] = name_dict[player_id]
        match["id"] = player_id
        match["score"] = best_matches[0]["score"]
    return match


def _create_player_team_dict(play_by_play):
    ids = []
    teams = []
    for pbp in play_by_play:
        ids.append(pbp.batter_id_br)
        ids.append(pbp.pitcher_id_br)
        teams.append(pbp.team_batting_id_br)
        teams.append(pbp.team_pitching_id_br)

    player_dict = dict(zip(ids, teams))
    for id in player_dict:
        if len(id) == 0:
            return None
    return player_dict


def _parse_in_game_substitutions(play_by_play_table, game_id, player_name_dict):
    row_numbers = play_by_play_table.xpath(_PBP_PLAYER_SUB_ROW_NUM_XPATH)
    if row_numbers is None:
        error = "No in game substitutions found in play-by-play table"
        return Result.Fail(error)
    sub_list = []
    for n in row_numbers:
        row_num = int(n)
        subs_xpath = Template(_T_PBP_PLAYER_SUB_XPATH).substitute(row=row_num)
        sub_descriptions = play_by_play_table.xpath(subs_xpath)
        if sub_descriptions is None:
            continue
        for i in range(len(sub_descriptions)):
            s = sub_descriptions[i].strip().replace("\xa0", " ")
            result = _parse_substitution_description(s)
            if result.failure:
                return result
            sub_dict = result.value
            substitution = BBRefInGameSubstitution(
                pbp_table_row_number=row_num,
                sub_description=sub_dict["description"],
                incoming_player_name=sub_dict["incoming_player_name"],
                outgoing_player_name=sub_dict["outgoing_player_name"],
                incoming_player_pos=sub_dict["incoming_player_pos"],
                outgoing_player_pos=sub_dict["outgoing_player_pos"],
                lineup_slot=sub_dict["lineup_slot"],
                sub_type=sub_dict["sub_type"],
                team_id="",
            )
            sub_list.append(substitution)

    result = {"sub_list": sub_list}
    return Result.Ok(result)


def _parse_substitution_description(sub_description):
    parsed_sub = {"description": sub_description}
    split, split2 = None, None
    pre_split = [s.strip() for s in sub_description.split("(change occurred mid-batter)")]
    sub_description = pre_split[0]
    if "running at second base to start the extra inning" in sub_description:
        des = "running at second base to start the extra inning"
        split = [s.strip() for s in sub_description.split(des)]
        parsed_sub["incoming_player_name"] = split[0]
        parsed_sub["incoming_player_pos"] = "PR"
        parsed_sub["outgoing_player_name"] = "N/A"
        parsed_sub["outgoing_player_pos"] = "N/A"
        parsed_sub["lineup_slot"] = 0
        parsed_sub["sub_type"] = "bat"
        remaining_description = split[1]
        if not remaining_description:
            return Result.Ok(parsed_sub)
    if "replaces" in sub_description:
        split = [s.strip() for s in sub_description.split("replaces")]
        parsed_sub["sub_type"] = "pitch"
    elif "and is now" in sub_description:
        split = [s.strip() for s in sub_description.split("pinch hit for")]
        parsed_sub["incoming_player_name"] = split[0]
        remaining_description = split[1]
        split2 = [s.strip() for s in remaining_description.split("and is now")]
        parsed_sub["incoming_player_pos"] = split[0]
        parsed_sub["outgoing_player_pos"] = split[0]
        parsed_sub["outgoing_player_name"] = "N/A"
        parsed_sub["lineup_slot"] = 0
        parsed_sub["sub_type"] = "bat"
        return Result.Ok(parsed_sub)
    elif "pinch hits for" in sub_description:
        parsed_sub["incoming_player_pos"] = "PH"
        parsed_sub["sub_type"] = "bat"
        split = [s.strip() for s in sub_description.split("pinch hits for")]
    elif "pinch runs for" in sub_description:
        parsed_sub["incoming_player_pos"] = "PR"
        parsed_sub["sub_type"] = "bat"
        split = [s.strip() for s in sub_description.split("pinch runs for")]
    elif "moves" in sub_description:
        split = [s.strip() for s in sub_description.split("moves")]
        parsed_sub["sub_type"] = "pitch"
    else:
        error = "Substitution description was in an unrecognized format. (Before first split)"
        return Result.Fail(error)

    if not split or len(split) != 2:
        error = "First split operation did not produce a list with length=2."
        return Result.Fail(error)
    parsed_sub["incoming_player_name"] = split[0]
    remaining_description = split[1]

    if "batting" in remaining_description:
        split2 = [s.strip() for s in remaining_description.split("batting")]
    elif "from" in remaining_description:
        match = _CHANGE_POS_REGEX.search(remaining_description)
        if match:
            match_dict = match.groupdict()
            parsed_sub["outgoing_player_name"] = parsed_sub["incoming_player_name"]
            parsed_sub["outgoing_player_pos"] = match_dict["old_pos"]
            parsed_sub["incoming_player_pos"] = match_dict["new_pos"]
            parsed_sub["lineup_slot"] = 0
            return Result.Ok(parsed_sub)
    elif remaining_description.endswith("pitching"):
        split2 = [s.strip() for s in remaining_description.split("pitching")]
        parsed_sub["outgoing_player_name"] = split2[0]
        parsed_sub["outgoing_player_pos"] = "P"
        parsed_sub["incoming_player_pos"] = "P"
        parsed_sub["lineup_slot"] = 0
        return Result.Ok(parsed_sub)
    else:
        error = "Substitution description was in an unrecognized format. (After first split)"
        return Result.Fail(error)

    if not split2 or len(split2) != 2:
        error = "Second split operation did not produce a list with length=2."
        return Result.Fail(error)
    remaining_description = split2[0]
    lineup_slot_str = split2[1]

    match = _NUM_REGEX.search(lineup_slot_str)
    if not match:
        error = "Failed to parse batting order number."
        return Result.Fail(error)
    parsed_sub["lineup_slot"] = int(match.group())

    split3 = None
    match = _POS_REGEX.search(remaining_description)
    if match:
        parsed_sub["outgoing_player_pos"] = match.group()[1:-1]
        split3 = [s.strip() for s in remaining_description.split(match.group())]
        if not split3 or len(split3) != 2:
            error = "Third split operation did not produce a list with length=2."
            return Result.Fail(error)
        parsed_sub["outgoing_player_name"] = split3[0]
        remaining_description = split3[1]
    if not remaining_description:
        return Result.Ok(parsed_sub)

    split4 = None
    if "pitching" in remaining_description:
        parsed_sub["incoming_player_pos"] = "P"
        parsed_sub["sub_type"] = "pitch"
        split4 = [s.strip() for s in remaining_description.split("pitching")]
        if not split4 or len(split4) != 2:
            error = "Fourth split operation did not produce a list with length=2."
            return Result.Fail(error)
        if "outgoing_player_name" not in parsed_sub:
            parsed_sub["outgoing_player_name"] = split4[0]
        remaining_description = split4[1]

    elif "playing" in remaining_description:
        split4 = [s.strip() for s in remaining_description.split("playing")]
        if not split4 or len(split4) != 2:
            error = "Fourth split operation did not produce a list with length=2."
            return Result.Fail(error)
        if "outgoing_player_name" not in parsed_sub:
            parsed_sub["outgoing_player_name"] = split4[0]
        remaining_description = split4[1]

        found_incoming_player_pos = False
        for pos in DEFENSE_POSITIONS:
            if pos in remaining_description:
                parsed_sub["incoming_player_pos"] = pos
                found_incoming_player_pos = True
                break
        if not found_incoming_player_pos:
            error = (
                "Unable to find incoming player position in remaining description:\n"
                f"remaining_description: {remaining_description}\n"
            )
            return Result.Fail(error)
    elif "pinch hit" in sub_description:
        parsed_sub["outgoing_player_name"] = remaining_description
        parsed_sub["outgoing_player_pos"] = "PH"
        parsed_sub["sub_type"] = "bat"
    elif "pinch run" in sub_description:
        parsed_sub["outgoing_player_name"] = remaining_description
        parsed_sub["outgoing_player_pos"] = "PR"
        parsed_sub["sub_type"] = "bat"

    if "outgoing_player_pos" not in parsed_sub:
        parsed_sub["outgoing_player_pos"] = parsed_sub["incoming_player_pos"]
    return Result.Ok(parsed_sub)


def _find_missing_pbp_events(summaries_begin, summaries_end, game_events, substitutions):
    row_ids_top_inning = list(summaries_begin.keys())
    row_ids_bot_inning = list(summaries_end.keys())
    row_ids_game_events = [int(event.pbp_table_row_number) for event in game_events]
    row_ids_substitutions = [sub.pbp_table_row_number for sub in substitutions]
    all_row_ids = set(row_ids_top_inning + row_ids_bot_inning + row_ids_game_events + row_ids_substitutions)
    first_row_num = min(all_row_ids)
    last_row_num = max(all_row_ids) + 1
    check_row_ids = set(range(first_row_num, last_row_num))
    missing_row_ids = check_row_ids - all_row_ids
    return list(missing_row_ids)


def _parse_missing_pbp_events(missing_row_ids, play_by_play_table, game_id):
    misc_events = []
    for row_num in missing_row_ids:
        missing_events_xpath = Template(_T_PBP_MISC_EVENT_XPATH).substitute(row=row_num)
        event_descriptions = play_by_play_table.xpath(missing_events_xpath)
        if event_descriptions is None:
            continue
        description = "".join(des.strip().replace("\xa0", " ") for des in event_descriptions)

        misc_event = BBRefPlayByPlayMiscEvent(pbp_table_row_number=row_num, description=description)
        misc_events.append(misc_event)
    return misc_events


def _create_innings_list(
    game_id,
    away_team_id,
    home_team_id,
    player_name_dict,
    player_id_match_log,
    summaries_begin,
    summaries_end,
    game_events,
    substitutions,
    misc_events,
):
    if len(summaries_begin) != len(summaries_end):
        error = "Begin inning and end inning summary list lengths do not match."
        return Result.Fail(error)

    start_boundaries = sorted(summaries_begin.keys())
    end_boundaries = sorted(summaries_end.keys())
    inning_number = 0
    innings_list = []
    for i in range(len(summaries_begin)):
        start_row = start_boundaries[i]
        end_row = end_boundaries[i]
        if is_even(i):
            inning_number += 1
            inning_label = f"t{inning_number}"
            inning_id = f"{game_id}_INN_TOP{inning_number:02d}"
        else:
            inning_label = f"b{inning_number}"
            inning_id = f"{game_id}_INN_BOT{inning_number:02d}"

        inning_events = [
            g for g in game_events if int(g.pbp_table_row_number) > start_row and int(g.pbp_table_row_number) < end_row
        ]
        for game_event in inning_events:
            game_event.inning_id = inning_id
            game_event.inning_label = inning_label

        inning_substitutions = [
            sub for sub in substitutions if sub.pbp_table_row_number > start_row and sub.pbp_table_row_number < end_row
        ]
        for sub in inning_substitutions:
            sub.inning_id = inning_id
            sub.inning_label = inning_label
            sub.team_id = _get_sub_team_id(inning_label, sub.sub_type, away_team_id, home_team_id)
            result = _get_sub_player_ids(
                sub.incoming_player_name,
                sub.outgoing_player_name,
                player_name_dict,
                sub.team_id,
            )
            if result.failure:
                return result
            id_dict = result.value
            sub.incoming_player_id_br = id_dict["incoming_player_id_br"]
            sub.outgoing_player_id_br = id_dict["outgoing_player_id_br"]
            if id_dict["player_id_match_log"]:
                player_id_match_log.extend(id_dict["player_id_match_log"])

        misc = [
            misc_event
            for misc_event in misc_events
            if misc_event.pbp_table_row_number > start_row and misc_event.pbp_table_row_number < end_row
        ]
        for me in misc:
            me.inning_id = inning_id
            me.inning_label = inning_label

        inning = BBRefHalfInning()
        inning.inning_id = inning_id
        inning.inning_label = inning_label
        inning.begin_inning_summary = summaries_begin[start_row]
        inning.end_inning_summary = summaries_end[end_row]
        inning.game_events = inning_events
        inning.substitutions = inning_substitutions
        inning.misc_events = misc

        match = _INNING_TOTALS_REGEX.search(summaries_end[end_row])
        if match:
            inning_totals = match.groupdict()
            inning.inning_total_runs = inning_totals["runs"]
            inning.inning_total_hits = inning_totals["hits"]
            inning.inning_total_errors = inning_totals["errors"]
            inning.inning_total_left_on_base = inning_totals["left_on_base"]
            inning.away_team_runs_after_inning = inning_totals["away_team_runs"]
            inning.home_team_runs_after_inning = inning_totals["home_team_runs"]

        innings_list.append(inning)
    return Result.Ok((innings_list, player_id_match_log))


def _get_sub_team_id(inning_label, sub_type, away_team_id, home_team_id):
    if inning_label.startswith("t") and "bat" in sub_type:
        return away_team_id
    elif inning_label.startswith("t") and "pitch" in sub_type:
        return home_team_id
    elif inning_label.startswith("b") and "bat" in sub_type:
        return home_team_id
    elif inning_label.startswith("b") and "pitch" in sub_type:
        return away_team_id
    return None


def _get_sub_player_ids(incoming_player_name, outgoing_player_name, player_name_dict, sub_team_id):
    player_id_match_log = []
    if outgoing_player_name != "N/A":
        match = _match_player_id(outgoing_player_name, sub_team_id, player_name_dict)
        outgoing_player_id_br = match["id"]
        if match["type"] != "Exact match":
            player_id_match_log.append(match)
    else:
        outgoing_player_id_br = "N/A"

    if incoming_player_name != "N/A":
        match = _match_player_id(incoming_player_name, sub_team_id, player_name_dict)
        incoming_player_id_br = match["id"]
        if match["type"] != "Exact match":
            player_id_match_log.append(match)
    else:
        incoming_player_id_br = "N/A"

    result = {
        "incoming_player_id_br": incoming_player_id_br,
        "outgoing_player_id_br": outgoing_player_id_br,
        "player_id_match_log": player_id_match_log,
    }
    return Result.Ok(result)



from dataclasses import dataclass

from vigorish.enums import PlayByPlayEvent


@dataclass
class BBRefPlayByPlayEvent:
    """Various numeric and string values that describe a plate appearance."""

    event_id: str = "0"
    inning_id: str = "0"
    inning_label: str = "0"
    pbp_table_row_number: str = "0"
    score: str = "0"
    outs_before_play: str = "0"
    runners_on_base: str = "0"
    pitch_sequence: str = "0"
    runs_outs_result: str = "0"
    team_batting_id_br: str = "0"
    team_pitching_id_br: str = "0"
    play_description: str = "0"
    pitcher_id_br: str = "0"
    batter_id_br: str = "0"
    play_index_url: str = "0"

    @property
    def event_type(self):
        return PlayByPlayEvent.AT_BAT

    def as_dict(self):
        """Convert game event values to a dictionary."""
        return {
            "__bbref_pbp_game_event__": True,
            "event_id": self.event_id,
            "inning_id": self.inning_id,
            "inning_label": self.inning_label,
            "pbp_table_row_number": int(self.pbp_table_row_number),
            "score": self.score,
            "outs_before_play": int(self.outs_before_play),
            "runners_on_base": self.runners_on_base,
            "pitch_sequence": self.pitch_sequence,
            "runs_outs_result": self.runs_outs_result,
            "team_batting_id_br": self.team_batting_id_br,
            "team_pitching_id_br": self.team_pitching_id_br,
            "play_description": self.play_description,
            "pitcher_id_br": self.pitcher_id_br,
            "batter_id_br": self.batter_id_br,
            "play_index_url": self.play_index_url,
        }



from datetime import datetime
from functools import cached_property

from events import Events

import vigorish.database as db
from vigorish.data.game_data import GameData
from vigorish.tasks.base import Task
from vigorish.util.dt_format_strings import DATE_ONLY_TABLE_ID
from vigorish.util.result import Result
from vigorish.util.string_helpers import get_bbref_team_id, validate_bbref_game_id


class AddToDatabaseTask(Task):
    def __init__(self, app):
        super().__init__(app)
        self._team_id_map = {}
        self.events = Events(
            (
                "find_all_games_eligible_for_import_start",
                "find_all_games_eligible_for_import_complete",
                "add_data_to_db_start",
                "add_data_to_db_progress",
                "add_data_to_db_complete",
            )
        )

    @cached_property
    def season_id_map(self):
        return db.Season.get_regular_season_map(self.db_session)

    @cached_property
    def player_id_map(self):
        return db.PlayerId.get_player_id_map(self.db_session)

    @cached_property
    def game_id_map(self):
        return db.GameScrapeStatus.get_game_id_map(self.db_session)

    @cached_property
    def pitch_app_id_map(self):
        return db.PitchAppScrapeStatus.get_pitch_app_id_map(self.db_session)

    def get_team_id_map_for_year(self, year):
        team_id_map_for_year = self._team_id_map.get(year)
        if not team_id_map_for_year:
            team_id_map_for_year = db.Team.get_team_id_map_for_year(self.db_session, year)
            self._team_id_map[year] = team_id_map_for_year
        return team_id_map_for_year

    def execute(self, year=None):
        return self.add_data_for_year(year) if year else self.add_all_data()

    def add_all_data(self):
        valid_years = [year for year, results in self.app.audit_report.items() if results["successful"]]
        for year in valid_years:
            self.add_data_for_year(year)
        return Result.Ok()

    def add_data_for_year(self, year):
        report_for_season = self.app.audit_report.get(year)
        if not report_for_season:
            return Result.Fail(f"Audit report could not be generated for MLB Season {year}")
        game_ids = report_for_season.get("successful")
        if not game_ids:
            error = f"No games for MLB Season {year} qualify to have PitchFx data imported."
            return Result.Fail(error)
        self.events.add_data_to_db_start(year, game_ids)
        for num, game_id in enumerate(game_ids, start=1):
            game_data = GameData(self.app, game_id)
            result = self.add_player_stats_to_database(game_id, game_data)
            if result.failure:
                return result
            result = self.add_pitchfx_to_database(game_data)
            if result.failure:
                return result
            self.events.add_data_to_db_progress(num, year, game_id)
        self.events.add_data_to_db_complete(year)
        return Result.Ok()

    def add_player_stats_to_database(self, game_id, game_data):
        game_status = db.GameScrapeStatus.find_by_bbref_game_id(self.db_session, game_id)
        if not game_status:
            error = f"Import aborted! Game status '{game_id}' not found in database"
            return Result.Fail(error)
        if not game_status.imported_bat_stats:
            self.add_bat_stats_to_database(game_id, game_data, game_status)
        if not game_status.imported_pitch_stats:
            self.add_pitch_stats_to_database(game_id, game_data, game_status)
        return Result.Ok()

    def add_bat_stats_to_database(self, game_id, game_data, game_status):
        for team_boxscore in game_data.bat_boxscore.values():
            for player_boxscore in team_boxscore.values():
                bat_stats_dict = game_data.get_bat_stats(player_boxscore["mlb_id"]).value
                if not bat_stats_dict:
                    continue
                bat_stats_dict["is_starter"] = player_boxscore["is_starter"]
                bat_stats_dict["bat_order"] = player_boxscore["bat_order"]
                bat_stats_dict["def_position"] = player_boxscore["def_position"]
                bat_stats = db.BatStats.from_dict(game_id, bat_stats_dict)
                bat_stats = self.update_player_stats_relationships(bat_stats)
                self.db_session.add(bat_stats)
        game_status.imported_bat_stats = 1
        self.db_session.commit()

    def add_pitch_stats_to_database(self, game_id, game_data, game_status):
        for mlb_id in game_data.pitch_stats_player_ids:
            pitch_stats_dict = game_data.get_pitch_app_stats(mlb_id).value
            pitch_stats = db.PitchStats.from_dict(game_id, pitch_stats_dict)
            pitch_stats = self.update_player_stats_relationships(pitch_stats)
            self.db_session.add(pitch_stats)
        game_status.imported_pitch_stats = 1
        self.db_session.commit()

    def update_player_stats_relationships(self, stats):
        game_date = self.get_game_date_from_bbref_game_id(stats.bbref_game_id)
        stats.player_id = self.player_id_map[stats.player_id_mlb]
        stats.player_team_id = self.get_team_id_map_for_year(game_date.year)[stats.player_team_id_bbref]
        stats.opponent_team_id = self.get_team_id_map_for_year(game_date.year)[stats.opponent_team_id_bbref]
        stats.season_id = self.season_id_map[game_date.year]
        stats.date_id = self.get_date_status_id_from_game_date(game_date)
        stats.game_status_id = self.game_id_map[stats.bbref_game_id]
        return stats

    def add_pitchfx_to_database(self, game_data):
        for pitch_app_id, pfx_dict_list in game_data.all_pitchfx.items():
            pitch_app = db.PitchAppScrapeStatus.find_by_pitch_app_id(self.db_session, pitch_app_id)
            if not pitch_app:
                error = f"Import aborted! Pitch app '{pitch_app_id}' not found in database"
                return Result.Fail(error)
            if pitch_app.imported_pitchfx:
                continue
            for pfx_dict in pfx_dict_list:
                pfx = db.PitchFx.from_dict(pfx_dict)
                pfx = self.update_pitchfx_relationships(pfx)
                self.db_session.add(pfx)
            pitch_app.imported_pitchfx = 1
        self.db_session.commit()
        return Result.Ok()

    def update_pitchfx_relationships(self, pfx):
        game_date = self.get_game_date_from_bbref_game_id(pfx.bbref_game_id)
        pitcher_team_id_br = get_bbref_team_id(pfx.pitcher_team_id_bb)
        opponent_team_id_br = get_bbref_team_id(pfx.opponent_team_id_bb)
        pfx.pitcher_id = self.player_id_map[pfx.pitcher_id_mlb]
        pfx.batter_id = self.player_id_map[pfx.batter_id_mlb]
        pfx.team_pitching_id = self.get_team_id_map_for_year(game_date.year)[pitcher_team_id_br]
        pfx.team_batting_id = self.get_team_id_map_for_year(game_date.year)[opponent_team_id_br]
        pfx.season_id = self.season_id_map[game_date.year]
        pfx.date_id = self.get_date_status_id_from_game_date(game_date)
        pfx.game_status_id = self.game_id_map[pfx.bbref_game_id]
        pfx.pitch_app_db_id = self.pitch_app_id_map[pfx.pitch_app_id]
        return pfx

    def get_game_date_from_bbref_game_id(self, bbref_game_id):
        game_date = validate_bbref_game_id(bbref_game_id).value["game_date"]
        return datetime(game_date.year, game_date.month, game_date.day)

    def get_date_status_id_from_game_date(self, game_date):
        return game_date.strftime(DATE_ONLY_TABLE_ID)
